from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.models import (
    CustomJS,
    MultiChoice,
    Div,
    HoverTool,
    Range1d,
    ColorBar,
    TapTool,
)
from bokeh.models import ColumnDataSource, LinearColorMapper, Button, SVGIcon

# from bokeh.models import Tabs, TabPanel,, Div, CrosshairTool, Span
from bokeh.palettes import Dark2_5
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bloqade.task.base import Report

import math

# import itertools
import numpy as np
from decimal import Decimal

# from typing import List


## =====================================================
# below are formatting IR data, and called by IR.


def format_report_data(report: "Report"):
    # return should be
    # data: List[ColumnDataSources],
    # List[str]:ch_names,
    # spmod_source:ColumnDataSources

    task_tid = report.dataframe.index.get_level_values("task_number").unique()
    task_tid = list(task_tid)

    counts = report.counts
    ryds = report.rydberg_densities()

    assert len(task_tid) == len(counts)

    cnt_sources = []
    ryd_sources = []
    for i, cnt_data in enumerate(counts):
        # bitstrings = list(
        #    "\n".join(textwrap.wrap(bitstring, 32)) for bitstring in cnt_data.keys()
        # )
        bitstrings = list(cnt_data.keys())
        bit_id = [f"[{x}]" for x in range(len(bitstrings))]

        cnts = list(cnt_data.values())
        tid = [i] * len(cnts)
        src = ColumnDataSource(
            data=dict(tid=tid, bitstrings=bitstrings, cnts=cnts, bit_id=bit_id)
        )

        rydens = list(ryds.iloc[i])
        tid = [i] * len(rydens)

        rsrc = ColumnDataSource(
            data=dict(
                tid=tid,
                sites=[f"{ds}" for ds in range(len(rydens))],
                ryds=rydens,
            )
        )
        cnt_sources.append(src)
        ryd_sources.append(rsrc)

    return cnt_sources, ryd_sources, report.metas, report.geos, report.name


def mock_data():
    import bloqade.task as tks

    cnt_sources = []
    ryd_sources = []
    metas = []
    geos = []

    # ===============================
    bitstrings = ["0010", "1101", "1111"]
    cnts = [4, 7, 5]
    tid = [0] * len(cnts)
    bit_id = [f"[{x}]" for x in range(len(bitstrings))]
    src = ColumnDataSource(
        data=dict(tid=tid, bitstrings=bitstrings, cnts=cnts, bit_id=bit_id)
    )
    cnt_sources.append(src)
    rsrc = ColumnDataSource(
        data=dict(
            tid=[0, 0, 0, 0],
            sites=["0", "1", "2", "3"],
            ryds=[0.55, 0.45, 0.2, 0.11],
        )
    )
    ryd_sources.append(rsrc)
    metas.append(dict(a=4, b=9, c=10))
    geos.append(
        tks.base.Geometry(
            sites=[
                (0.0e-6, 0.0e-6),
                (0.0e-6, 0.5e-6),
                (0.5e-6, 0.3e-6),
                (0.6e-6, 0.7e-6),
            ],
            filling=[1, 1, 1, 0],
        )
    )

    # ===============================
    bitstrings = ["0101", "1101", "1110", "1111"]
    cnts = [4, 7, 5, 10]
    tid = [1] * len(cnts)
    bit_id = [f"[{x}]" for x in range(len(bitstrings))]
    src = ColumnDataSource(
        data=dict(tid=tid, bitstrings=bitstrings, cnts=cnts, bit_id=bit_id)
    )
    cnt_sources.append(src)
    rsrc = ColumnDataSource(
        data=dict(
            tid=[1, 1, 1, 1],
            sites=["0", "1", "2", "3"],
            ryds=[0.33, 0.67, 0.25, 0.34],
        )
    )
    ryd_sources.append(rsrc)
    metas.append(dict(a=10, b=9.44, c=10.3))
    geos.append(
        tks.base.Geometry(
            sites=[
                (0.0e-6, 0.1e-6),
                (0.66e-6, 0.5e-6),
                (0.3e-6, 0.3e-6),
                (0.5e-6, 0.7e-6),
            ],
            filling=[1, 0, 1, 0],
        )
    )

    return cnt_sources, ryd_sources, metas, geos, "Mock"


def plot_register_ryd_dense(geo, ryds):
    """obtain a figure object from the atom arrangement."""
    xs_filled, ys_filled, labels_filled, density_filled = [], [], [], []
    xs_vacant, ys_vacant, labels_vacant, density_vacant = [], [], [], []
    x_min = np.inf
    x_max = -np.inf
    y_min = np.inf
    y_max = -np.inf
    for idx, location_info in enumerate(zip(geo.sites, geo.filling, ryds)):
        (x, y), filling, density = location_info
        x = float(Decimal(str(x)) * Decimal("1e6"))  # convert to um
        y = float(Decimal(str(y)) * Decimal("1e6"))  # convert to um
        x_min = min(x, x_min)
        y_min = min(y, y_min)
        x_max = max(x, x_max)
        y_max = max(y, y_max)
        if filling:
            xs_filled.append(x)
            ys_filled.append(y)
            labels_filled.append(idx)
            density_filled.append(density)
        else:
            xs_vacant.append(x)
            ys_vacant.append(y)
            labels_vacant.append(idx)
            density_vacant.append(density)

    if len(geo.sites) > 0:
        length_scale = max(y_max - y_min, x_max - x_min, 1)
    else:
        length_scale = 1

    source_filled = ColumnDataSource(
        data=dict(
            _x=xs_filled, _y=ys_filled, _labels=labels_filled, _ryd=density_filled
        )
    )
    source_vacant = ColumnDataSource(
        data=dict(
            _x=xs_vacant, _y=ys_vacant, _labels=labels_vacant, _ryd=density_vacant
        )
    )

    hover = HoverTool()
    hover.tooltips = [
        ("(x,y)", "(@_x, @_y)"),
        ("index: ", "@_labels"),
        ("ryd density: ", "@_ryd"),
    ]

    color_mapper = LinearColorMapper(palette="Magma256", low=min(ryds), high=max(ryds))

    # specify that we want to map the colors to the y values,
    # this could be replaced with a list of colors
    ##p.scatter(x,y,color={'field': 'y', 'transform': color_mapper})

    ## remove box_zoom since we don't want to change the scale

    p = figure(
        width=500,
        height=400,
        tools="wheel_zoom,reset, undo, redo, pan",
        toolbar_location="above",
        title="rydberg density",
    )
    p.x_range = Range1d(x_min - 1, x_min + length_scale + 1)
    p.y_range = Range1d(y_min - 1, y_min + length_scale + 1)

    p.circle(
        "_x",
        "_y",
        source=source_filled,
        radius=0.035 * length_scale,
        fill_alpha=1,
        line_color="black",
        color={"field": "_ryd", "transform": color_mapper},
    )
    p.circle(
        "_x",
        "_y",
        source=source_vacant,
        radius=0.035 * length_scale,
        fill_alpha=1,
        # color="grey",
        line_color="black",
        color={"field": "_ryd", "transform": color_mapper},
        line_width=0.2 * length_scale,
    )

    color_bar = ColorBar(
        color_mapper=color_mapper,
        label_standoff=12,
        border_line_color=None,
        location=(0, 0),
    )

    p.xaxis.axis_label = "(um)"
    p.add_layout(color_bar, "right")
    p.add_tools(hover)

    return p


def plot_register_bits(geo):
    """obtain a figure object from the atom arrangement."""
    # xs_filled, ys_filled, labels_filled, density_filled = [], [], [], []
    # xs_vacant, ys_vacant, labels_vacant, density_vacant = [], [], [], []
    xs = []
    ys = []
    bits = []
    labels = []

    x_min = np.inf
    x_max = -np.inf
    y_min = np.inf
    y_max = -np.inf
    for idx, location_info in enumerate(zip(geo.sites, geo.filling)):
        (x, y), filling = location_info
        x = float(Decimal(str(x)) * Decimal("1e6"))  # convert to um
        y = float(Decimal(str(y)) * Decimal("1e6"))  # convert to um
        x_min = min(x, x_min)
        y_min = min(y, y_min)
        x_max = max(x, x_max)
        y_max = max(y, y_max)

        xs.append(x)
        ys.append(y)
        bits.append(0)
        labels.append(idx)

    if len(geo.sites) > 0:
        length_scale = max(y_max - y_min, x_max - x_min, 1)
    else:
        length_scale = 1

    source = ColumnDataSource(data=dict(_x=xs, _y=ys, _bits=bits, _labels=labels))

    hover = HoverTool()
    hover.tooltips = [
        ("(x,y)", "(@_x, @_y)"),
        ("index: ", "@_labels"),
        ("state: ", "@_bits"),
    ]

    color_mapper = LinearColorMapper(palette="Magma256", low=0, high=1)

    # specify that we want to map the colors to the y values,
    # this could be replaced with a list of colors
    ##p.scatter(x,y,color={'field': 'y', 'transform': color_mapper})

    ## remove box_zoom since we don't want to change the scale

    p = figure(
        width=400,
        height=400,
        tools="wheel_zoom,reset, undo, redo, pan",
        toolbar_location="above",
        title="reg state",
    )
    p.x_range = Range1d(x_min - 1, x_min + length_scale + 1)
    p.y_range = Range1d(y_min - 1, y_min + length_scale + 1)

    p.circle(
        "_x",
        "_y",
        source=source,
        radius=0.035 * length_scale,
        fill_alpha=1,
        line_color="black",
        color={"field": "_bits", "transform": color_mapper},
        name="reg",
    )

    p.xaxis.axis_label = "(um)"
    p.add_tools(hover)

    return p


def report_visual(cnt_sources, ryd_sources, metas, geos, name):
    options = [f"task {cnt}" for cnt in range(len(cnt_sources))]

    figs = []
    # select = Select(title="Select Task", options=[])
    multi_choice = MultiChoice(options=[])

    if len(options):
        color1 = Dark2_5[1]
        # color2 = Dark2_5[1]
        for taskname, tsrc, trydsrc, meta, geo in zip(
            options, cnt_sources, ryd_sources, metas, geos
        ):
            content = "<p> Assignments: </p>"
            for var, num in meta.items():
                content += f"<p>{var} = {num}</p>"

            div = Div(
                text=content, width=100, height=400, styles={"overflow-y": "scroll"}
            )

            xrng = [list(int(i) for i in x) for x in tsrc.data["bitstrings"]]

            # print(xrng)

            p = figure(
                x_range=tsrc.data["bit_id"],
                height=400,
                width=400,
                title=f"{taskname}",
                # toolbar_location=None,
                tools="xwheel_zoom,reset, box_zoom, xpan",
            )
            bar_rend = p.vbar(
                x="bit_id", top="cnts", source=tsrc, width=0.9, color=color1
            )

            p.xgrid.grid_line_color = None
            p.xaxis.major_label_orientation = math.pi / 4
            p.y_range.start = 0
            p.yaxis.axis_label = "Counts"

            hov_tool = HoverTool()
            hov_tool.tooltips = [
                ("counts: ", "@cnts"),
                # ("bitstrings:\n", "@bitstrings"),
            ]
            p.add_tools(hov_tool)

            tap = TapTool(renderers=[bar_rend])
            p.add_tools(tap)
            p.toolbar.active_tap = tap

            preg = plot_register_bits(geo)
            # get render obj:
            reg_obj = None
            for rd in preg.renderers:
                if rd.name == "reg":
                    reg_obj = rd
                    break

            cb = CustomJS(
                args=dict(
                    tsrc=tsrc,
                    xrng=xrng,
                    bitsrc=reg_obj.data_source,
                    reg_obj=reg_obj,
                    preg=preg,
                ),
                code="""
                            var sel_bar_i = tsrc.selected.indices[0];
                            bitsrc.data['_bits'] = xrng[sel_bar_i];
                            preg.title.text = "reg state: [" + sel_bar_i + "]";
                            bitsrc.change.emit();
                            reg_obj.change.emit();
                            preg.change.emit();
                          """,
            )
            tsrc.selected.js_on_change("indices", cb)

            # pryd = figure(
            #    x_range=trydsrc.data["sites"],
            #    height=400,
            #    width=300,
            #    # toolbar_location=None,
            #    tools="xwheel_zoom,reset,box_zoom,xpan",
            # )

            # pryd.vbar(x="sites", top="ryds", source=trydsrc, width=0.5, color=color2)

            # pryd.xgrid.grid_line_color = None
            # pryd.y_range.start = 0
            # pryd.yaxis.axis_label = "Rydberg density"
            # pryd.xaxis.major_label_orientation = math.pi / 4
            # pryd.xaxis.axis_label = "site"

            # hov_tool = HoverTool()
            # hov_tool.tooltips = [("density: ", "@ryds")]
            # pryd.add_tools(hov_tool)

            # pgeo = plot_register(geo)
            # print(geo,  trydsrc.data["ryds"])
            pgeo = plot_register_ryd_dense(geo, trydsrc.data["ryds"])

            figs.append(row(div, p, preg, pgeo, name=taskname))
            figs[-1].visible = False

        # Create a dropdown menu to select between the two graphs
        figs[0].visible = True
        multi_choice = MultiChoice(value=[options[0]], options=options)

        multi_choice.js_on_change(
            "value",
            CustomJS(
                args=dict(figs=figs, options=options, nelem=len(options)),
                code="""
                        const vals = this.value
                        for(let i=0;i<nelem;i++){
                            figs[i].visible=false;
                            for(let j=0;j<vals.length;j++){
                                if(vals[j]==options[i]){
                                    figs[i].visible=true;
                                    break;
                                }
                            }
                        }
                    """,
            ),
        )

    # headline = row(Div(text="Report: " + name), bt)
    headline = Div(text="Report: " + name)

    bt = Button(label="Bloqade", icon=SVGIcon(svg=bloqadeICON()))
    bt.js_on_click(
        CustomJS(args=dict(url="https://bloqade.quera.com"), code="window.open(url)")
    )

    bt2 = Button(label="QuEra", icon=SVGIcon(svg=queraICON()))
    bt2.js_on_click(
        CustomJS(args=dict(url="https://www.quera.com/"), code="window.open(url)")
    )

    return column(column(headline, row(bt, bt2)), column(multi_choice, column(*figs)))


def queraICON():
    return """
        <svg
    id="Layer_2"
    viewBox="0 0 154.96999 188.87825"
    version="1.1"
    sodipodi:docname="logo black.svg"
    width="154.96999"
    height="188.87825"
    inkscape:export-filename="QElogo.svg"
    inkscape:export-xdpi="96"
    inkscape:export-ydpi="96"
    xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
    xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
    xmlns="http://www.w3.org/2000/svg"
    xmlns:svg="http://www.w3.org/2000/svg">
    <defs
        id="defs98" />
    <sodipodi:namedview
        id="namedview98"
        pagecolor="#ffffff"
        bordercolor="#000000"
        borderopacity="0.25"
        inkscape:showpageshadow="2"
        inkscape:pageopacity="0.0"
        inkscape:pagecheckerboard="0"
        inkscape:deskcolor="#d1d1d1" />
    <g
        id="g98"
        transform="translate(-90.150002,-1.4817416)">
        <circle
        cx="154.87"
        cy="178.75"
        r="11.61"
        id="circle7" />
        <circle
        cx="-177.98538"
        cy="126.21368"
        r="11.61"
        transform="rotate(-89.82)"
        id="circle8" />
        <circle
        cx="184.17"
        cy="178.75"
        r="11.61"
        id="circle9" />
        <circle
        cx="154.87"
        cy="13.46"
        r="11.61"
        id="circle10" />
        <circle
        cx="-12.696921"
        cy="125.69522"
        r="11.61"
        transform="rotate(-89.82)"
        id="circle11" />
        <circle
        cx="184.17"
        cy="13.46"
        r="11.61"
        id="circle12" />
        <circle
        cx="101.76"
        cy="159.53"
        r="11.61"
        id="circle13" />
        <circle
        cx="101.76"
        cy="126.13"
        r="11.61"
        id="circle14" />
        <circle
        cx="101.76"
        cy="95.919998"
        r="11.61"
        id="circle15" />
        <circle
        cx="101.76"
        cy="65.699997"
        r="11.61"
        id="circle16" />
        <circle
        cx="101.76"
        cy="35.490002"
        r="11.61"
        id="circle17" />
        <circle
        cx="154.87"
        cy="121.1"
        r="11.61"
        id="circle86" />
        <circle
        cx="181.08"
        cy="140.32001"
        r="11.61"
        id="circle87" />
        <circle
        cx="233.50999"
        cy="178.75"
        r="11.61"
        id="circle88" />
        <circle
        cx="207.42"
        cy="126.13"
        r="11.61"
        id="circle89" />
        <circle
        cx="207.42"
        cy="95.919998"
        r="11.61"
        id="circle90" />
        <circle
        cx="207.42"
        cy="65.699997"
        r="11.61"
        id="circle91" />
        <circle
        cx="207.42"
        cy="35.490002"
        r="11.61"
        id="circle92" />
        <circle
        cx="207.42"
        cy="159.53"
        r="11.61"
        id="circle93" />
        <circle
        cx="102.38"
        cy="126.13"
        r="11.61"
        id="circle94" />
        <circle
        cx="102.38"
        cy="95.919998"
        r="11.61"
        id="circle95" />
        <circle
        cx="102.38"
        cy="65.699997"
        r="11.61"
        id="circle96" />
        <circle
        cx="102.38"
        cy="35.490002"
        r="11.61"
        id="circle97" />
        <circle
        cx="102.38"
        cy="159.53"
        r="11.61"
        id="circle98" />
    </g>
    </svg>
    """


def bloqadeICON():
    return """
    <!-- Generator: Adobe Illustrator 27.0.0, SVG Export Plug-In .
        SVG Version: 6.00 Build 0)  -->

    <svg
    version="1.1"
    id="Layer_1"
    x="0px"
    y="0px"
    viewBox="0 0 120 125"
    xml:space="preserve"
    sodipodi:docname="logo.svg"
    width="120"
    height="125"
    inkscape:version="1.3 (0e150ed, 2023-07-21)"
    xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
    xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
    xmlns="http://www.w3.org/2000/svg"
    xmlns:svg="http://www.w3.org/2000/svg"><defs
    id="defs16"><radialGradient
        id="SVGID_1_"
        cx="284.23792"
        cy="40.978001"
        r="70.431198"
        gradientUnits="userSpaceOnUse">&#10;				<stop
    offset="0"
    style="stop-color:#FFFFFF"
    id="stop7" />&#10;				<stop
    offset="0.1264"
    style="stop-color:#FBFAFF"
    id="stop8" />&#10;				<stop
    offset="0.2679"
    style="stop-color:#F1EDFF"
    id="stop9" />&#10;				<stop
    offset="0.4168"
    style="stop-color:#E0D6FF"
    id="stop10" />&#10;				<stop
    offset="0.5707"
    style="stop-color:#C7B7FF"
    id="stop11" />&#10;				<stop
    offset="0.7287"
    style="stop-color:#A88FFF"
    id="stop12" />&#10;				<stop
    offset="0.8875"
    style="stop-color:#825EFF"
    id="stop13" />&#10;				<stop
    offset="1"
    style="stop-color:#6437FF"
    id="stop14" />&#10;			</radialGradient></defs><sodipodi:namedview
    id="namedview16"
    pagecolor="#ffffff"
    bordercolor="#000000"
    borderopacity="0.25"
    inkscape:showpageshadow="2"
    inkscape:pageopacity="0.0"
    inkscape:pagecheckerboard="0"
    inkscape:deskcolor="#d1d1d1"
    inkscape:zoom="1.3545752"
    inkscape:cx="308.95296"
    inkscape:cy="72.347407"
    inkscape:window-width="1456"
    inkscape:window-height="803"
    inkscape:window-x="260"
    inkscape:window-y="138"
    inkscape:window-maximized="0"
    inkscape:current-layer="Layer_1" />&#10;<style
    type="text/css"
    id="style1">&#10;	.st0{fill:url(#SVGID_1_);}&#10;</style>&#10;&#10;<g
    id="g14"
    transform="matrix(0.97296819,0,0,1,-234.63028,-11.63275)">&#10;
    <path
    d="m 356,136.42 -15.67,-13.98 c -9.94,7.75 -22.57,12.47 -37.06,
    12.47 -35.88,0 -59.97,-28.13 -59.97,
    -60.64 v -0.34 c 0,-32.51 24.42,-60.98 60.3,
    -60.98 35.88,0 59.97,28.13 59.97,60.64 v 0.34 c 0,
    14.66 -5.22,28.64 -13.98,39.42 l 15.5,13.14 z m -25.1,
    -22.57 -22.07,-18.87 8.93,-9.94 22.07,20.04 c 6.4,
    -8.42 9.94,-19.03 9.94,-30.82 v -0.34 c 0,-26.78 -19.54,
    -48.68 -46.49,-48.68 -26.95,0 -46.15,
    21.56 -46.15,48.34 v 0.34 c 0,26.79 19.54,48.68 46.49,
    48.68 10.44,0 19.87,-3.19 27.28,-8.75 z"
    id="path7" />&#10;			<radialGradient
    id="radialGradient16"
    cx="284.23792"
    cy="40.978001"
    r="70.431198"
    gradientUnits="userSpaceOnUse">&#10;				<stop
    offset="0"
    style="stop-color:#FFFFFF"
    id="stop1" />&#10;				<stop
    offset="0.1264"
    style="stop-color:#FBFAFF"
    id="stop2" />&#10;				<stop
    offset="0.2679"
    style="stop-color:#F1EDFF"
    id="stop3" />&#10;				<stop
    offset="0.4168"
    style="stop-color:#E0D6FF"
    id="stop4" />&#10;				<stop
    offset="0.5707"
    style="stop-color:#C7B7FF"
    id="stop5" />&#10;				<stop
    offset="0.7287"
    style="stop-color:#A88FFF"
    id="stop6" />&#10;				<stop
    offset="0.8875"
    style="stop-color:#825EFF"
    id="stop15" />&#10;				<stop
    offset="1"
    style="stop-color:#6437FF"
    id="stop16" />&#10;			</radialGradient>&#10;			<circle
    class="st0"
    cx="303.78"
    cy="74.68"
    r="49.439999"
    id="circle14"
    style="fill:url(#SVGID_1_)" />&#10;		</g></svg>
    """



import bloqade.ir.control.waveform as waveform
from bloqade.ir.control.waveform import Record
from bloqade.ir.visitor.waveform import WaveformVisitor

from typing import Dict, Tuple, List, Union
from pydantic.dataclasses import dataclass
from bisect import bisect_left, bisect_right
import numbers
from decimal import Decimal


@dataclass(frozen=True)
class PiecewiseLinear:
    times: List[Decimal]
    values: List[Decimal]

    def eval(self, time: Decimal) -> Decimal:
        if time >= self.times[-1]:
            return self.values[-1]

        elif time <= self.times[0]:
            return self.values[0]
        else:
            index = bisect_right(self.times, time) - 1

            m = (self.values[index + 1] - self.values[index]) / (
                self.times[index + 1] - self.times[index]
            )
            t = time - self.times[index]
            b = self.values[index]

            return m * t + b

    def slice(self, start_time: Decimal, stop_time: Decimal) -> "PiecewiseLinear":
        start_time = (
            Decimal(str(start_time))
            if not isinstance(start_time, Decimal)
            else start_time
        )
        stop_time = (
            Decimal(str(stop_time)) if not isinstance(stop_time, Decimal) else stop_time
        )

        if start_time == stop_time:
            return PiecewiseLinear(
                [Decimal(0.0), Decimal(0.0)], [Decimal(0.0), Decimal(0.0)]
            )

        start_index = bisect_left(self.times, start_time)
        stop_index = bisect_left(self.times, stop_time)
        start_value = self.eval(start_time)
        stop_value = self.eval(stop_time)

        if start_time == self.times[start_index]:
            if stop_time == self.times[start_index]:
                absolute_times = list(self.times[start_index : stop_index + 1])
                values = list(self.values[start_index : stop_index + 1])
            else:
                absolute_times = self.times[start_index:stop_index] + [stop_time]
                values = self.values[start_index:stop_index] + [stop_value]
        else:
            if stop_time == self.times[stop_index]:
                absolute_times = [start_time] + self.times[start_index : stop_index + 1]
                values = [start_value] + self.values[start_index : stop_index + 1]
            else:
                absolute_times = (
                    [start_time] + self.times[start_index:stop_index] + [stop_time]
                )
                values = (
                    [start_value] + self.values[start_index:stop_index] + [stop_value]
                )

        return PiecewiseLinear([time - start_time for time in absolute_times], values)

    def append(self, other: "PiecewiseLinear"):
        assert self.values[-1] == other.values[0]

        return PiecewiseLinear(
            times=self.times + [time + self.times[-1] for time in other.times[1:]],
            values=self.values + other.values[1:],
        )


class PiecewiseLinearCodeGen(WaveformVisitor):
    def __init__(self, assignments: Dict[str, Union[numbers.Real, List[numbers.Real]]]):
        self.assignments = assignments
        self.times = []
        self.values = []

    @staticmethod
    def check_continiuity(left, right):
        if left != right:
            diff = abs(left - right)
            raise ValueError(
                f"discontinuity with a jump of {diff} found when compiling to "
                "piecewise linear."
            )

    def append_timeseries(self, start, stop, duration):
        if len(self.times) == 0:
            self.times = [Decimal(0), duration]
            self.values = [start, stop]
        else:
            self.check_continiuity(self.values[-1], start)

            self.times.append(duration + self.times[-1])
            self.values.append(stop)

    def visit_linear(self, ast: waveform.Linear) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        start = ast.start(**self.assignments)
        stop = ast.stop(**self.assignments)
        self.append_timeseries(start, stop, duration)

    def visit_constant(
        self, ast: waveform.Constant
    ) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.duration(**self.assignments)
        value = ast.value(**self.assignments)
        self.append_timeseries(value, value, duration)

    def visit_poly(self, ast: waveform.Poly) -> Tuple[List[Decimal], List[Decimal]]:
        order = len(ast.coeffs) - 1
        duration = ast.duration(**self.assignments)

        if len(ast.coeffs) == 0:
            start = Decimal(0)
            stop = Decimal(0)
        elif len(ast.coeffs) == 1:
            start = ast.coeffs[0](**self.assignments)
            stop = start
        elif len(ast.coeffs) == 2:
            start = ast.coeffs[0](**self.assignments)
            stop = start + ast.coeffs[1](**self.assignments) * duration
        else:
            raise ValueError(
                "Failed to compile Waveform to piecewise linear,"
                f"found Polynomial of order {order}."
            )

        self.append_timeseries(start, stop, duration)

    def visit_negative(
        self, ast: waveform.Negative
    ) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)
        self.values = [-value for value in self.values]

    def visit_scale(self, ast: waveform.Scale) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)
        scaler = ast.scalar(**self.assignments)
        self.values = [scaler * value for value in self.values]

    def visit_slice(self, ast: waveform.Slice) -> Tuple[List[Decimal], List[Decimal]]:
        duration = ast.waveform.duration(**self.assignments)
        if ast.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = ast.interval.start(**self.assignments)

        if ast.interval.stop is None:
            stop_time = duration
        else:
            stop_time = ast.interval.stop(**self.assignments)

        if start_time < 0:
            raise ValueError((f"start time for slice {start_time} is smaller than 0."))

        if stop_time > duration:
            raise ValueError(
                (f"end time for slice {stop_time} is larger than duration {duration}.")
            )

        if stop_time < start_time:
            raise ValueError(
                (
                    f"end time for slice {stop_time} is smaller than "
                    f"starting value for slice {start_time}."
                )
            )

        pwl = PiecewiseLinearCodeGen(self.assignments).emit(ast.waveform)
        new_pwl = pwl.slice(start_time, stop_time)

        self.times = new_pwl.times
        self.values = new_pwl.values

    def visit_append(self, ast: waveform.Append) -> Tuple[List[Decimal], List[Decimal]]:
        pwl = PiecewiseLinearCodeGen(self.assignments).emit(ast.waveforms[0])

        for sub_expr in ast.waveforms[1:]:
            new_pwl = PiecewiseLinearCodeGen(self.assignments).emit(sub_expr)

            # skip instructions with duration=0
            if new_pwl.times[-1] == Decimal(0):
                continue

            self.check_continiuity(pwl.values[-1], new_pwl.values[0])
            pwl = pwl.append(new_pwl)

        self.times = pwl.times
        self.values = pwl.values

    def visit_sample(self, ast: waveform.Sample) -> Tuple[List[Decimal], List[Decimal]]:
        if ast.interpolation is not waveform.Interpolation.Linear:
            raise ValueError(
                "Failed to compile waveform to piecewise linear, "
                f"found piecewise {ast.interpolation.value} interpolation."
            )

        self.times, self.values = ast.samples(**self.assignments)

    def visit_record(self, ast: Record) -> Tuple[List[Decimal], List[Decimal]]:
        self.visit(ast.waveform)

    def emit(self, ast: waveform.Waveform) -> PiecewiseLinear:
        self.visit(ast)
        return PiecewiseLinear(self.times, self.values)



from bloqade.builder.backend.quera import QuEraService
from bloqade.builder.backend.braket import BraketService
from bloqade.builder.backend.bloqade import BloqadeService


class BackendRoute(QuEraService, BraketService, BloqadeService):
    """
    Specify the backend to run your program on via a string
    (versus more formal builder syntax) of specifying the vendor/product first
    (Bloqade/Braket) and narrowing it down
    (e.g: ...device("quera.aquila") versus ...quera.aquila())
    - You can pass the following arguments:
        - `"braket.aquila"`
        - `"braket.local_emulator"`
        - `"bloqade.python"`
        - `"bloqade.julia"`

    """

    def device(self, name: str, *args, **kwargs):
        if name == "quera.aquila":
            dev = self.quera.aquila
        elif name == "braket.aquila":
            dev = self.braket.aquila
        elif name == "braket.local_emulator":
            dev = self.braket.local_emulator
        elif name == "bloqade.python":
            dev = self.bloqade.python
        elif name == "bloqade.julia":
            dev = self.bloqade.julia
        else:
            raise ValueError(f"Unknown device: {name}")
        return dev(*args, **kwargs)



from collections import OrderedDict
from pydantic.dataclasses import dataclass
from beartype import beartype
from beartype.typing import Optional, Tuple
from bloqade.builder.typing import LiteralType

from bloqade.ir.routine.base import RoutineBase, __pydantic_dataclass_config__
from bloqade.submission.braket import BraketBackend
from bloqade.task.batch import LocalBatch, RemoteBatch
from bloqade.task.braket_simulator import BraketEmulatorTask
from bloqade.task.braket import BraketTask


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BraketServiceOptions(RoutineBase):
    def device(self, device_arn: str) -> "BraketHardwareRoutine":
        backend = BraketBackend(device_arn=device_arn)
        return BraketHardwareRoutine(self.source, self.circuit, self.params, backend)

    def aquila(self) -> "BraketHardwareRoutine":
        backend = BraketBackend(
            device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
        )
        return BraketHardwareRoutine(self.source, self.circuit, self.params, backend)

    def local_emulator(self):
        return BraketLocalEmulatorRoutine(self.source, self.circuit, self.params)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BraketHardwareRoutine(RoutineBase):
    backend: BraketBackend

    def _compile(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
    ) -> RemoteBatch:
        ## fall passes here ###
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
        from bloqade.ir.analysis.assignment_scan import AssignmentScan
        from bloqade.codegen.hardware.quera import AHSCodegen

        capabilities = self.backend.get_capabilities()

        circuit, params = self.circuit, self.params
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            # TODO: Replace these two steps with:
            # task_ir, parallel_decoder = BraketCodeGen().emit(shots, final_circuit)
            result = AHSCodegen(shots, capabilities=capabilities).emit(final_circuit)
            metadata = {**params.static_params, **record_params}
            task_ir = result.quera_task_ir.discretize(capabilities)
            tasks[task_number] = BraketTask(
                None,
                self.backend,
                task_ir,
                metadata,
                result.parallel_decoder,
                None,
            )

        batch = RemoteBatch(source=self.source, tasks=tasks, name=name)

        return batch

    @beartype
    def run_async(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, and run_async to Braket.

        Note:
            This is async.

        Args:
            shots (int): number of shots
            args (Tuple): Values of the parameter defined in `args`, defaults to ()
            name (str | None): custom name of the batch, defaults to None
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """

        batch = self._compile(shots, args, name)
        batch._submit(shuffle, **kwargs)
        return batch

    @beartype
    def run(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ) -> RemoteBatch:
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, run_async to Braket,
        and wait until the results are coming back.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args (Tuple): additional arguments
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """

        batch = self.run_async(shots, args, name, shuffle, **kwargs)
        batch.pull()
        return batch

    @beartype
    def __call__(
        self,
        *args: LiteralType,
        shots: int = 1,
        name: Optional[str] = None,
        shuffle: bool = False,
        **kwargs,
    ):
        """
        Compile to a RemoteBatch, which contain
        Braket backend specific tasks, run_async to Braket,
        and wait until the results are coming back.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args: additional arguments for args variables.
            name (str): custom name of the batch
            shuffle (bool): shuffle the order of jobs

        Return:
            RemoteBatch

        """
        return self.run(shots, args, name, shuffle, **kwargs)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BraketLocalEmulatorRoutine(RoutineBase):
    def _compile(
        self, shots: int, args: Tuple[LiteralType, ...] = (), name: Optional[str] = None
    ) -> LocalBatch:
        ## fall passes here ###
        from bloqade.ir import ParallelRegister
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
        from bloqade.codegen.hardware.quera import AHSCodegen
        from bloqade.ir.analysis.assignment_scan import AssignmentScan

        circuit, params = self.circuit, self.params
        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        if isinstance(circuit.register, ParallelRegister):
            raise TypeError(
                "Parallelization of atom arrangements is not supported for "
                "local emulation."
            )

        tasks = OrderedDict()

        for task_number, batch_params in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_params).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            # TODO: Replace these two steps with:
            # task_ir, _ = BraketCodeGen().emit(shots, final_circuit)
            result = AHSCodegen(shots).emit(final_circuit)

            metadata = {**params.static_params, **record_params}
            tasks[task_number] = BraketEmulatorTask(
                result.braket_task_ir,
                metadata,
                None,
            )

        batch = LocalBatch(source=self.source, tasks=tasks, name=name)

        return batch

    @beartype
    def run(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        **kwargs,
    ) -> LocalBatch:
        """
        Compile to a LocalBatch, and run.
        The LocalBatch contain tasks to run on local emulator.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args: additional arguments for args variables.
            multiprocessing (bool): enable multi-process
            num_workers (int): number of workers to run the emulator

        Return:
            LocalBatch

        """

        batch = self._compile(shots, args, name)
        batch._run(multiprocessing=multiprocessing, num_workers=num_workers, **kwargs)
        return batch

    @beartype
    def __call__(
        self,
        *args: LiteralType,
        shots: int = 1,
        name: Optional[str] = None,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        **kwargs,
    ):
        """
        Compile to a LocalBatch, and run.
        The LocalBatch contain tasks to run on local emulator.

        Note:
            This is sync, and will wait until remote results
            finished.

        Args:
            shots (int): number of shots
            args: additional arguments for args variables.
            multiprocessing (bool): enable multi-process
            num_workers (int): number of workers to run the emulator

        Return:
            LocalBatch

        """
        return self.run(
            shots,
            args,
            name,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            **kwargs,
        )



from bloqade.builder.coupling import Rydberg, Hyperfine


class Drive:
    @property
    def rydberg(self) -> Rydberg:
        """
        Address the Rydberg level coupling in your program.

        - Next possible steps to build your program are specifying the
          [`Rabi`][bloqade.builder.field.Rabi] field or
          [`Detuning`][bloqade.builder.field.Detuning] field.
            - `...rydberg.rabi`: for Rabi field
            - `...rydberg.detuning`: for Detuning field
        - In the absence of a field you the value is set to zero by default.
        """
        return Rydberg(self)

    @property
    def hyperfine(self) -> Hyperfine:
        """
        Address the Hyperfine level coupling in your program.

        - Next possible steps to build your program are specifying the
          [`Rabi`][bloqade.builder.field.Rabi] field or
          [`Detuning`][bloqade.builder.field.Detuning] field.
            - `...hyperfine.rabi`: for Rabi field
            - `...hyperfine.detuning`: for Detuning field
        - In the absence of a field you the value is set to zero by default.

        """
        return Hyperfine(self)



from ..scalar import Interval
from ..tree_print import Printer
from .field import Field
from typing import List
from pydantic.dataclasses import dataclass
from bloqade.visualization import get_pulse_figure
from bloqade.visualization import display_ir

__all__ = [
    "Pulse",
    "NamedPulse",
    "FieldName",
    "rabi",
    "detuning",
]


@dataclass(frozen=True)
class FieldName:
    def print_node(self):
        return self.__class__.__name__

    def children(self):
        return []

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __eq__(self, other):
        return self.__class__ == other.__class__


@dataclass(frozen=True)
class RabiFrequencyAmplitude(FieldName):
    def print_node(self):
        return "RabiFrequencyAmplitude"


@dataclass(frozen=True)
class RabiFrequencyPhase(FieldName):
    def print_node(self):
        return "RabiFrequencyPhase"


@dataclass(frozen=True)
class Detuning(FieldName):
    def print_node(self):
        return "Detuning"


class RabiRouter:
    def __init__(self) -> None:
        self.amplitude = RabiFrequencyAmplitude()
        self.phase = RabiFrequencyPhase()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def print_node(self):
        return "RabiRouter"

    def children(self):
        return {"Amplitude": self.amplitude, "Phase": self.phase}


rabi = RabiRouter()
detuning = Detuning()


@dataclass
class PulseExpr:
    """
    ```bnf
    <expr> ::= <pulse>
      | <append>
      | <slice>
      | <named>
    ```
    """

    def append(self, other: "PulseExpr") -> "PulseExpr":
        return PulseExpr.canonicalize(Append([self, other]))

    def slice(self, interval: Interval) -> "PulseExpr":
        return PulseExpr.canonicalize(Slice(self, interval))

    @staticmethod
    def canonicalize(expr: "PulseExpr") -> "PulseExpr":
        # TODO: update canonicalization rules for appending pulses

        if isinstance(expr, Append):
            new_pulses = []
            for pulse in expr.pulses:
                if isinstance(pulse, Append):
                    new_pulses += pulse.pulses
                else:
                    new_pulses.append(pulse)

            new_pulses = list(map(PulseExpr.canonicalize, new_pulses))
            return Append(new_pulses)
        else:
            return expr

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def _get_data(self, **assigments):
        return NotImplementedError

    def figure(self, **assignments):
        return NotImplementedError

    def show(self, **assignments):
        return NotImplementedError


@dataclass
class Append(PulseExpr):
    """
    ```bnf
    <append> ::= <expr>+
    ```
    """

    pulses: List[PulseExpr]

    def print_node(self):
        return "Append"

    def children(self):
        return self.pulses


@dataclass(init=False)
class Pulse(PulseExpr):
    """
    ```bnf
    <pulse> ::= (<field name> <field>)+
    ```
    """

    fields: dict[FieldName, Field]

    def __init__(self, field_pairs):
        fields = dict()
        for k, v in field_pairs.items():
            if isinstance(v, Field):
                fields[k] = v
            elif isinstance(v, dict):
                fields[k] = Field(v)
            else:
                raise TypeError(f"Expected Field or dict, got {type(v)}")
        self.fields = fields

    def print_node(self):
        return "Pulse"

    def children(self):
        # annotated children
        return {k.print_node(): v for k, v in self.fields.items()}

    def _get_data(self, **assigments):
        return None, self.fields

    def figure(self, **assignments):
        return get_pulse_figure(self, **assignments)

    def show(self, **assignments):
        """
        Interactive visualization of the Pulse

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the Pulse

        """
        display_ir(self, assignments)


@dataclass
class NamedPulse(PulseExpr):
    name: str
    pulse: PulseExpr

    def print_node(self):
        return "NamedPulse"

    def children(self):
        return {"Name": self.name, "Pulse": self.pulse}

    def _get_data(self, **assigments):
        return self.name, self.pulse.value

    def figure(self, **assignments):
        return get_pulse_figure(self, **assignments)

    def show(self, **assignments):
        display_ir(self, assignments)


@dataclass
class Slice(PulseExpr):
    pulse: PulseExpr
    interval: Interval

    def print_node(self):
        return "Slice"

    def children(self):
        return {"Pulse": self.pulse, "Interval": self.interval}



from bloqade.codegen.hardware.waveform import WaveformCodeGen
from bloqade.codegen.hardware.field import FieldCodeGen
from bloqade.codegen.hardware.pulse import PulseCodeGen

from bloqade.ir.pulse import RabiFrequencyAmplitude, RabiFrequencyPhase
from bloqade.ir.scalar import cast
from bloqade.ir.waveform import Constant, Linear, Append
from bloqade.ir.field import Field, Global
from bloqade.ir.sequence import Pulse, RydbergLevelCoupling


initial_detuning = cast("initial_detuning")
final_detuning = cast("final_detuning")
up_time = cast("up_time")
anneal_time = cast("anneal_time")

detuning_wf = Append(
    [
        Constant(initial_detuning, up_time),
        Linear(initial_detuning, final_detuning, anneal_time),
        Constant(final_detuning, up_time),
    ]
)

assignments = dict(
    initial_detuning=-15, final_detuning=10, up_time=0.1, anneal_time=4.0
)

wf_codegen = WaveformCodeGen(10, assignments, field_name=RabiFrequencyAmplitude())

times, values = wf_codegen.emit(detuning_wf[0.05:4.15])

print(times)
print(values)

field = Field({Global: detuning_wf})

field_codegen = FieldCodeGen(
    10,
    assignments,
    field_name=RabiFrequencyAmplitude(),
)
quera_field = field_codegen.emit(field)

print(quera_field)


field = Field({Global: detuning_wf})
field_codegen = FieldCodeGen(
    10,
    assignments,
    field_name=RabiFrequencyPhase(),
)
try:
    quera_field = field_codegen.emit(field)
except NotImplementedError:
    pass

pulse = Pulse({RabiFrequencyAmplitude(): field})


pulse_codegen = PulseCodeGen(10, assignments, level_coupling=RydbergLevelCoupling())
rydberg_hamiltonian = pulse_codegen.emit(pulse)
print(rydberg_hamiltonian)



from beartype.typing import List, Dict, Union, TYPE_CHECKING
from bloqade.builder.typing import LiteralType, ParamType
from bloqade.ir.scalar import Variable

if TYPE_CHECKING:
    from bloqade.builder.assign import Assign, BatchAssign, ListAssign
    from bloqade.builder.parallelize import Parallelize
    from bloqade.builder.args import Args


class AddArgs:
    def args(self, args_list: List[Union[str, Variable]]) -> "Args":
        from bloqade.builder.args import Args

        return Args(args_list, self)


class Assignable:
    def assign(self, **assignments) -> "Assign":
        """
        Assign values to variables declared previously in the program.

        This is reserved for variables that should only take single values OR
        for spatial modulations that were created with `.scale(str)` in which case
        you can pass in a list. This is the ONLY circumstance in which multiple
        values are allowed.

        ### Usage Examples:
        ```
        # define geometry
        >>> reg = bloqade.start
        ...       .add_position([(0,0),(1,1),(2,2),(3,3)])
        # define variables in program
        >>> seq = reg.rydberg.detuning.uniform
        ...       .linear(start="ival",stop=1,duration="span_time")
        # assign values to variables
        >>> seq = seq.assign(span_time = 0.5, ival = 0.0)
        ```

        - You can now:
            - `...assign(assignments).bloqade`: select the bloqade local
                emulator backend
            - `...assign(assignments).braket`: select braket local emulator or
                QuEra hardware
            - `...assign(assignments).device(specifier_string)`: select backend
                by specifying a string
        - Assign multiple values to a single variable for a parameter sweep:
            - `...assign(assignments).batch_assign(assignments)`:
        - Parallelize the program register, duplicating the geometry and waveform
            sequence to take advantage of all available
          space/qubits on the QPU:
            - `...assign(assignments).parallelize(cluster_spacing)`
        - Defer value assignment of certain variables to runtime:
            - `...assign(assignments).args([previously_defined_vars])`

        """
        from bloqade.builder.assign import Assign

        return Assign(assignments, parent=self)


class BatchAssignable:
    def batch_assign(
        self,
        __batch_params: List[Dict[str, ParamType]] = [],
        **assignments: List[ParamType],
    ) -> Union["BatchAssign", "ListAssign"]:
        """

        Assign multiple values to a single variable to create a parameter sweep.

        Bloqade automatically handles the multiple programs this would generate
        and treats it as object with unified results for easy post-processing.

        NOTE: if you assign multiple values to multiple variables in your program,
        the values must be of the same length. Bloqade will NOT do a Cartesian product
        (e.g. if "var1" is assigned [1,2,3] and "var2" is assigned [4,5,6] then the
        resulting programs will have assignments [1,4], [2,5], [3,6]).

        ### Usage Example:
        ```
        >>> reg = start.add_position([(0,0), (0, "atom_distance")])
        >>> prog = reg.rydberg.rabi.amplitude.uniform.constant("value", 5.0)
        >>> var_assigned_prog = prog.batch_assign(value = [1.0, 2.0, 3.0],
        atom_distance = [1.0, 2.0, 3.0])
        ```

        - Next steps are:
            - `...batch_assign(assignments).bloqade`: select the bloqade
                local emulator backend
            - `...batch_assign(assignments).braket`: select braket local
            emulator or QuEra hardware
            - `...batch_assign(assignments).device(specifier_string)`: select
            backend by specifying a string
        - Parallelize the program register, duplicating the geometry and waveform
          sequence to take advantage of all available
          space/qubits on the QPU:
            - `...batch_assign(assignments).parallelize(cluster_spacing)`
        - Defer value assignment of certain variables to runtime:
            - `...batch_assign(assignments).args([previously_defined_vars])`

        """
        from bloqade.builder.assign import BatchAssign, ListAssign

        if len(__batch_params) > 0 and assignments:
            raise ValueError("batch_params and assignments cannot be used together.")

        if len(__batch_params) > 0:
            return ListAssign(__batch_params, parent=self)
        else:
            return BatchAssign(assignments, parent=self)


class Parallelizable:
    def parallelize(self, cluster_spacing: LiteralType) -> "Parallelize":
        """

        Parallelize the current problem (register and sequence) by duplicating
        the geometry to take advantage of all available space/qubits on hardware.

        The singular argument lets you specify how far apart the clusters
        should be in micrometers.

        ### Usage Example:
        ```
        >>> reg = start.add_position((0,0)).rydberg.rabi.uniform.amplitude
        .constant(1.0, 1.0)
        # copy-paste the geometry and waveforms
        >>> parallelized_prog = reg.parallelize(24)
        ```

        - Your next steps are:
             `...parallelize(cluster_spacing).bloqade`: select the bloqade
                local emulator backend
             `...parallelize(cluster_spacing).braket`: select braket
                local emulator or QuEra hardware on the cloud
             `...parallelize(cluster_spacing).device(specifier_string)`: select
                backend by specifying a string

        """
        from bloqade.builder.parallelize import Parallelize

        return Parallelize(cluster_spacing, self)



from collections import OrderedDict
from bloqade.ir.routine.base import RoutineBase, __pydantic_dataclass_config__
from bloqade.builder.typing import LiteralType
from bloqade.task.batch import LocalBatch
from beartype import beartype
from beartype.typing import Optional, Tuple
from pydantic.dataclasses import dataclass


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BloqadeServiceOptions(RoutineBase):
    def python(self):
        return BloqadePythonRoutine(self.source, self.circuit, self.params)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BloqadePythonRoutine(RoutineBase):
    def _compile(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        blockade_radius: LiteralType = 0.0,
        cache_matrices: bool = False,
    ) -> LocalBatch:
        from bloqade.ir.analysis.assignment_scan import AssignmentScan
        from bloqade.codegen.common.assign_variables import AssignAnalogCircuit
        from bloqade.codegen.emulator_ir import EmulatorProgramCodeGen
        from bloqade.emulate.codegen.hamiltonian import CompileCache
        from bloqade.task.bloqade import BloqadeTask

        circuit, params = self.circuit, self.params

        circuit = AssignAnalogCircuit(params.static_params).visit(circuit)

        if cache_matrices:
            matrix_cache = CompileCache()
        else:
            matrix_cache = None

        tasks = OrderedDict()
        for task_number, batch_param in enumerate(params.batch_assignments(*args)):
            record_params = AssignmentScan(batch_param).emit(circuit)
            final_circuit = AssignAnalogCircuit(record_params).visit(circuit)
            emulator_ir = EmulatorProgramCodeGen(blockade_radius=blockade_radius).emit(
                final_circuit
            )
            metadata = {**params.static_params, **record_params}
            tasks[task_number] = BloqadeTask(shots, emulator_ir, metadata, matrix_cache)

        return LocalBatch(self.source, tasks, name)

    @beartype
    def run(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        blockade_radius: float = 0.0,
        interaction_picture: bool = False,
        cache_matrices: bool = False,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        solver_name: str = "dop853",
        atol: float = 1e-14,
        rtol: float = 1e-7,
        nsteps: int = 2_147_483_647,
    ) -> LocalBatch:
        """Run the current program using bloqade python backend

        Args:
            shots (int): number of shots after running state vector simulation
            args (Tuple[Real, ...], optional): The values for parameters defined
            in `args`. Defaults to ().
            name (Optional[str], optional): Name to give this run. Defaults to None.
            blockade_radius (float, optional): Use the Blockade subspace given a
            particular radius. Defaults to 0.0.
            interaction_picture (bool, optional): Use the interaction picture when
            solving schrodinger equation. Defaults to False.
            cache_matrices (bool, optional): Reuse previously evaluated matrcies when
            possible. Defaults to False.
            multiprocessing (bool, optional): Use multiple processes to process the
            batches. Defaults to False.
            num_workers (Optional[int], optional): Number of processes to run with
            multiprocessing. Defaults to None.
            solver_name (str, optional): Which SciPy Solver to use. Defaults to
            "dop853".
            atol (float, optional): Absolute tolerance for ODE solver. Defaults to
            1e-14.
            rtol (float, optional): Relative tolerance for adaptive step in ODE solver.
            Defaults to 1e-7.
            nsteps (int, optional): Maximum number of steps allowed per integration
            step. Defaults to 2_147_483_647, the maximum value.

        Raises:
            ValueError: Cannot use multiprocessing and cache_matrices at the same time.

        Returns:
            LocalBatch: Batch of local tasks that have been executed.
        """
        if multiprocessing and cache_matrices:
            raise ValueError(
                "Cannot use multiprocessing and cache_matrices at the same time."
            )

        compile_options = dict(
            shots=shots,
            args=args,
            name=name,
            blockade_radius=blockade_radius,
            cache_matrices=cache_matrices,
        )

        solver_options = dict(
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )

        batch = self._compile(**compile_options)
        batch._run(**solver_options)

        return batch

    @beartype
    def __call__(
        self,
        *args: LiteralType,
        shots: int = 1,
        name: Optional[str] = None,
        blockade_radius: float = 0.0,
        interaction_picture: bool = False,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        cache_matrices: bool = False,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
    ):
        options = dict(
            shots=shots,
            args=args,
            name=name,
            blockade_radius=blockade_radius,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            cache_matrices=cache_matrices,
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )
        return self.run(**options)



import braket.ir.ahs as braket_ir
from braket.ahs.pattern import Pattern
from braket.timings import TimeSeries
from braket.ahs.atom_arrangement import AtomArrangement, SiteType
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.driving_field import DrivingField
from braket.ahs.shifting_field import ShiftingField
from braket.ahs.field import Field

from braket.task_result import AnalogHamiltonianSimulationTaskResult
from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraTaskStatusCode,
    QuEraShotResult,
    QuEraShotStatusCode,
)

from bloqade.submission.ir.task_specification import (
    QuEraTaskSpecification,
    GlobalField,
    LocalField,
)
from typing import Tuple, Union, List
from pydantic import BaseModel
from decimal import Decimal


class BraketTaskSpecification(BaseModel):
    nshots: int
    program: braket_ir.Program


def to_braket_time_series(times: List[Decimal], values: List[Decimal]) -> TimeSeries:
    time_series = TimeSeries()
    for time, value in zip(times, values):
        time_series.put(time, value)

    return time_series


def to_braket_field(quera_field: Union[GlobalField, LocalField]) -> Field:
    if isinstance(quera_field, GlobalField):
        times = quera_field.times
        values = quera_field.values
        time_series = to_braket_time_series(times, values)
        return Field(pattern="uniform", time_series=time_series)
    elif isinstance(quera_field, LocalField):
        times = quera_field.times
        values = quera_field.values
        pattern = quera_field.lattice_site_coefficients
        time_series = to_braket_time_series(times, values)
        pattern = Pattern(pattern)
        return Field(pattern=pattern, time_series=time_series)
    else:
        raise TypeError


def extract_braket_program(quera_task_ir: QuEraTaskSpecification):
    lattice = quera_task_ir.lattice

    rabi_amplitude = (
        quera_task_ir.effective_hamiltonian.rydberg.rabi_frequency_amplitude.global_
    )
    rabi_phase = (
        quera_task_ir.effective_hamiltonian.rydberg.rabi_frequency_phase.global_
    )
    global_detuning = quera_task_ir.effective_hamiltonian.rydberg.detuning.global_
    local_detuning = quera_task_ir.effective_hamiltonian.rydberg.detuning.local

    register = AtomArrangement()
    for site, filled in zip(lattice.sites, lattice.filling):
        site_type = SiteType.FILLED if filled == 1 else SiteType.VACANT
        register.add(site, site_type)

    hamiltonian = DrivingField(
        amplitude=to_braket_field(rabi_amplitude),
        phase=to_braket_field(rabi_phase),
        detuning=to_braket_field(global_detuning),
    )

    if local_detuning:
        hamiltonian = hamiltonian + ShiftingField(to_braket_field(local_detuning))

    return AnalogHamiltonianSimulation(
        register=register,
        hamiltonian=hamiltonian,
    )


def to_braket_task(
    quera_task_ir: QuEraTaskSpecification,
) -> Tuple[int, AnalogHamiltonianSimulation]:
    braket_ahs_program = extract_braket_program(quera_task_ir)
    return quera_task_ir.nshots, braket_ahs_program


def to_braket_task_ir(quera_task_ir: QuEraTaskSpecification) -> BraketTaskSpecification:
    nshots, braket_ahs_program = to_braket_task(quera_task_ir)
    return BraketTaskSpecification(nshots=nshots, program=braket_ahs_program.to_ir())


def from_braket_task_results(
    braket_task_results: AnalogHamiltonianSimulationTaskResult,
) -> QuEraTaskResults:
    shot_outputs = []
    for measurement in braket_task_results.measurements:
        shot_outputs.append(
            QuEraShotResult(
                shot_status=QuEraShotStatusCode.Completed,
                pre_sequence=list(measurement.pre_sequence),
                post_sequence=list(measurement.post_sequence),
            )
        )

    return QuEraTaskResults(
        task_status=QuEraTaskStatusCode.Completed, shot_outputs=shot_outputs
    )


def from_braket_status_codes(braket_message: str) -> QuEraTaskStatusCode:
    if braket_message == str("QUEUED"):
        return QuEraTaskStatusCode.Enqueued
    else:
        return QuEraTaskStatusCode(braket_message.lower().capitalize())



from bloqade.serialize import Serializer
from bloqade.task.base import Report
from bloqade.task.quera import QuEraTask
from bloqade.task.braket import BraketTask
from bloqade.task.braket_simulator import BraketEmulatorTask
from bloqade.task.bloqade import BloqadeTask

from bloqade.builder.base import Builder

from bloqade.submission.ir.task_results import (
    QuEraShotStatusCode,
    QuEraTaskStatusCode,
    QuEraTaskResults,
)

# from bloqade.submission.base import ValidationError

from beartype.typing import Union, Optional, Dict, Any, List
from beartype import beartype
from collections import OrderedDict
from itertools import product
import traceback
import datetime
import sys
import os
import warnings
import pandas as pd
import numpy as np
from dataclasses import dataclass, field


class Serializable:
    def json(self, **options) -> str:
        """
        Serialize the object to JSON string.

        Return:
            JSON string

        """
        from bloqade import dumps

        return dumps(self, **options)


@dataclass
@Serializer.register
class LocalBatch(Serializable):
    source: Optional[Builder]
    tasks: OrderedDict[int, Union[BraketEmulatorTask, BloqadeTask]]
    name: Optional[str] = None

    def report(self) -> Report:
        """
        Generate analysis report base on currently
        completed tasks in the LocalBatch.

        Return:
            Report

        """

        ## this potentially can be specialize/disatch
        ## offline
        index = []
        data = []
        metas = []
        geos = []

        for task_number, task in self.tasks.items():
            geometry = task.geometry
            perfect_sorting = "".join(map(str, geometry.filling))
            parallel_decoder = geometry.parallel_decoder

            if parallel_decoder:
                cluster_indices = parallel_decoder.get_cluster_indices()
            else:
                cluster_indices = {(0, 0): list(range(len(perfect_sorting)))}

            shot_iter = filter(
                lambda shot: shot.shot_status == QuEraShotStatusCode.Completed,
                task.result().shot_outputs,
            )

            for shot, (cluster_coordinate, cluster_index) in product(
                shot_iter, cluster_indices.items()
            ):
                pre_sequence = "".join(
                    map(
                        str,
                        (shot.pre_sequence[index] for index in cluster_index),
                    )
                )

                post_sequence = np.asarray(
                    [shot.post_sequence[index] for index in cluster_index],
                    dtype=np.int8,
                )

                pfc_sorting = "".join(
                    [perfect_sorting[index] for index in cluster_index]
                )

                key = (
                    task_number,
                    cluster_coordinate,
                    pfc_sorting,
                    pre_sequence,
                )

                index.append(key)
                data.append(post_sequence)

            metas.append(task.metadata)
            geos.append(task.geometry)

        index = pd.MultiIndex.from_tuples(
            index, names=["task_number", "cluster", "perfect_sorting", "pre_sequence"]
        )

        df = pd.DataFrame(data, index=index)
        df.sort_index(axis="index")

        rept = None
        if self.name is None:
            rept = Report(df, metas, geos, "Local")
        else:
            rept = Report(df, metas, geos, self.name)

        return rept

    @beartype
    def rerun(
        self, multiprocessing: bool = False, num_workers: Optional[int] = None, **kwargs
    ):
        """
        Rerun all the tasks in the LocalBatch.

        Return:
            Report

        """

        return self._run(
            multiprocessing=multiprocessing, num_workers=num_workers, **kwargs
        )

    def _run(
        self, multiprocessing: bool = False, num_workers: Optional[int] = None, **kwargs
    ):
        if multiprocessing:
            from concurrent.futures import ProcessPoolExecutor as Pool

            with Pool(max_workers=num_workers) as pool:
                futures = OrderedDict()
                for task_number, task in enumerate(self.tasks.values()):
                    futures[task_number] = pool.submit(task.run, **kwargs)

                for task_number, future in futures.items():
                    self.tasks[task_number] = future.result()

        else:
            if num_workers is not None:
                raise ValueError(
                    "num_workers is only used when multiprocessing is enabled."
                )
            for task in self.tasks.values():
                task.run(**kwargs)

        return self


@LocalBatch.set_serializer
def _serialize(obj: LocalBatch) -> Dict[str, Any]:
    return {
        "source": None,
        "tasks": [(k, v) for k, v in obj.tasks.items()],
        "name": obj.name,
    }


@LocalBatch.set_deserializer
def _deserializer(d: Dict[str, Any]) -> LocalBatch:
    d["tasks"] = OrderedDict(d["tasks"])
    return LocalBatch(**d)


@dataclass
@Serializer.register
class TaskError(Serializable):
    exception_type: str
    stack_trace: str


@dataclass
@Serializer.register
class BatchErrors(Serializable):
    task_errors: OrderedDict[int, TaskError] = field(
        default_factory=lambda: OrderedDict([])
    )

    @beartype
    def print_errors(self, task_indices: Union[List[int], int]) -> str:
        return str(self.get_errors(task_indices))

    @beartype
    def get_errors(self, task_indices: Union[List[int], int]):
        return BatchErrors(
            task_errors=OrderedDict(
                [
                    (task_index, self.task_errors[task_index])
                    for task_index in task_indices
                    if task_index in self.task_errors
                ]
            )
        )

    def __str__(self) -> str:
        output = ""
        for task_index, task_error in self.task_errors.items():
            output += (
                f"Task {task_index} failed to submit with error: "
                f"{task_error.exception_type}\n"
                f"{task_error.stack_trace}"
            )

        return output


@BatchErrors.set_serializer
def _serialize(self: BatchErrors) -> Dict[str, List]:
    return {
        "task_errors": [
            (task_number, task_error)
            for task_number, task_error in self.task_errors.items()
        ]
    }


@BatchErrors.set_deserializer
def _deserialize(obj: dict) -> BatchErrors:
    return BatchErrors(task_errors=OrderedDict(obj["task_errors"]))


# this class get collection of tasks
# basically behaves as a psudo queuing system
# the user only need to store this objecet
@dataclass
@Serializer.register
class RemoteBatch(Serializable):
    source: Builder
    tasks: Union[OrderedDict[int, QuEraTask], OrderedDict[int, BraketTask]]
    name: Optional[str] = None

    class SubmissionException(Exception):
        pass

    @property
    def total_nshots(self):
        """
        Total number of shots of all tasks in the RemoteBatch

        Return:
            number of shots

        """
        nshots = 0
        for task in self.tasks.values():
            nshots += task.task_ir.nshots
        return nshots

    def cancel(self) -> "RemoteBatch":
        """
        Cancel all the tasks in the Batch.

        Return:
            self

        """
        # cancel all jobs
        for task in self.tasks.values():
            task.cancel()

        return self

    def fetch(self) -> "RemoteBatch":
        """
        Fetch the tasks in the Batch.

        Note:
            Fetching will update the status of tasks,
            and only pull the results for those tasks
            that have completed.

        Return:
            self

        """
        # online, non-blocking
        # pull the results only when its ready
        for task in self.tasks.values():
            task.fetch()

        return self

    def pull(self) -> "RemoteBatch":
        """
        Pull results of the tasks in the Batch.

        Note:
            Pulling will pull the results for the tasks.
            If a given task(s) has not been completed, wait
            until it finished.

        Return:
            self
        """
        # online, blocking
        # pull the results. if its not ready, hanging
        for task in self.tasks.values():
            task.pull()

        return self

    def __repr__(self) -> str:
        return str(self.tasks_metric())

    def tasks_metric(self) -> pd.DataFrame:
        """
        Get current tasks status metric

        Return:
            dataframe with ["task id", "status", "shots"]

        """
        # [TODO] more info on current status
        # offline, non-blocking
        tid = []
        data = []
        for int, task in self.tasks.items():
            tid.append(int)

            dat = [None, None, None]
            dat[0] = task.task_id
            if task.task_result_ir is not None:
                dat[1] = task.task_result_ir.task_status.name
                dat[2] = task.task_ir.nshots
            data.append(dat)

        return pd.DataFrame(data, index=tid, columns=["task ID", "status", "shots"])

    def remove_invalid_tasks(self) -> "RemoteBatch":
        """
        Create a RemoteBatch object that
        contain tasks from current Batch,
        with all Unaccepted tasks removed.

        Return:
            RemoteBatch

        """
        return self.remove_tasks("Unaccepted")

        # return RemoteBatch(new_tasks, name=self.name)

    @beartype
    def resubmit(self, shuffle_submit_order: bool = True) -> "RemoteBatch":
        """
        Resubmit all the tasks in the RemoteBatch

        Return:
            self

        """
        # online, non-blocking
        self._submit(shuffle_submit_order, force=True)
        return self

    def _submit(
        self, shuffle_submit_order: bool = True, ignore_submission_error=False, **kwargs
    ) -> "RemoteBatch":
        from bloqade import save

        # online, non-blocking
        if shuffle_submit_order:
            submission_order = np.random.permutation(list(self.tasks.keys()))
        else:
            submission_order = list(self.tasks.keys())

        # submit tasks in random order but store them
        # in the original order of tasks.
        # futures = OrderedDict()

        ## upon submit() should validate for Both backends
        ## and throw errors when fail.
        errors = BatchErrors()
        shuffled_tasks = OrderedDict()
        for task_index in submission_order:
            task = self.tasks[task_index]
            shuffled_tasks[task_index] = task
            try:
                task.submit(**kwargs)
            except BaseException as error:
                # record the error in the error dict
                errors.task_errors[int(task_index)] = TaskError(
                    exception_type=error.__class__.__name__,
                    stack_trace=traceback.format_exc(),
                )

                task.task_result_ir = QuEraTaskResults(
                    task_status=QuEraTaskStatusCode.Unaccepted
                )

        self.tasks = shuffled_tasks  # permute order using dump way

        if len(errors.task_errors) > 0:
            time_stamp = datetime.datetime.now()

            if "win" in sys.platform:
                time_stamp = str(time_stamp).replace(":", "~")

            if self.name:
                future_file = f"{self.name}-partial-batch-future-{time_stamp}.json"
                error_file = f"{self.name}-partial-batch-errors-{time_stamp}.json"
            else:
                future_file = f"partial-batch-future-{time_stamp}.json"
                error_file = f"partial-batch-errors-{time_stamp}.json"

            cwd = os.getcwd()
            # cloud_batch_result.save_json(future_file, indent=2)
            # saving ?

            save(errors, error_file)
            save(self, future_file)

            if ignore_submission_error:
                warnings.warn(
                    "One or more error(s) occured during submission, please see "
                    "the following files for more information:\n"
                    f"  - {os.path.join(cwd, future_file)}\n"
                    f"  - {os.path.join(cwd, error_file)}\n",
                    RuntimeWarning,
                )
            else:
                raise RemoteBatch.SubmissionException(
                    str(errors)
                    + "\n"
                    + "One or more error(s) occured during submission, please see "
                    "the following files for more information:\n"
                    f"  - {os.path.join(cwd, future_file)}\n"
                    f"  - {os.path.join(cwd, error_file)}\n"
                )

        else:
            # TODO: think about if we should automatically save successful submissions
            #       as well.
            pass

    @beartype
    def get_tasks(self, *status_codes: str) -> "RemoteBatch":
        """
        Get Tasks with specify status_codes.

        Return:
            RemoteBatch

        """
        # offline:
        st_codes = [QuEraTaskStatusCode(x) for x in status_codes]

        new_task_results = OrderedDict()
        for task_number, task in self.tasks.items():
            if task.task_result_ir.task_status in st_codes:
                new_task_results[task_number] = task

        return RemoteBatch(self.source, new_task_results, name=self.name)

    @beartype
    def remove_tasks(self, *status_codes: str) -> "RemoteBatch":
        """
        Remove Tasks with specify status_codes.

        Return:
            RemoteBatch

        """
        # offline:

        st_codes = [QuEraTaskStatusCode(x) for x in status_codes]

        new_results = OrderedDict()
        for task_number, task in self.tasks.items():
            if task.task_result_ir.task_status in st_codes:
                continue

            new_results[task_number] = task

        return RemoteBatch(self.source, new_results, self.name)

    def get_failed_tasks(self) -> "RemoteBatch":
        """
        Create a RemoteBatch object that
        contain failed tasks from current Batch.

        failed tasks with following status codes:

        1. Failed
        2. Unaccepted

        Return:
            RemoteBatch

        """
        # statuses that are in a state that are
        # completed because of an error
        statuses = ["Failed", "Unaccepted"]
        return self.get_tasks(*statuses)

    def remove_failed_tasks(self) -> "RemoteBatch":
        """
        Create a RemoteBatch object that
        contain tasks from current Batch,
        with failed tasks removed.

        failed tasks with following status codes:

        1. Failed
        2. Unaccepted

        Return:
            RemoteBatch

        """
        # statuses that are in a state that will
        # not run going forward because of an error
        statuses = ["Failed", "Unaccepted"]
        return self.remove_tasks(*statuses)

    def get_finished_tasks(self) -> "RemoteBatch":
        """
        Create a RemoteBatch object that
        contain finished tasks from current Batch.

        Tasks consider finished with following status codes:

        1. Failed
        2. Unaccepted
        3. Completed
        4. Partial
        5. Cancelled

        Return:
            RemoteBatch

        """
        # statuses that are in a state that will
        # not run going forward for any reason
        statuses = ["Completed", "Failed", "Unaccepted", "Partial", "Cancelled"]
        return self.remove_tasks(*statuses)

    def get_completed_tasks(self) -> "RemoteBatch":
        """
        Create a RemoteBatch object that
        contain completed tasks from current Batch.

        Tasks consider completed with following status codes:

        1. Completed
        2. Partial

        Return:
            RemoteBatch

        """
        statuses = [
            "Completed",
            "Partial",
        ]
        return self.get_tasks(*statuses)

    def report(self) -> "Report":
        """
        Generate analysis report base on currently
        completed tasks in the RemoteBatch.

        Return:
            Report

        """
        ## this potentially can be specialize/disatch
        ## offline
        index = []
        data = []
        metas = []
        geos = []

        for task_number, task in self.tasks.items():
            ## fliter not existing results tasks:
            if (task.task_id is None) or (not task._result_exists()):
                continue

            ## filter has result but is not correctly completed.
            if not task.task_result_ir.task_status == QuEraTaskStatusCode.Completed:
                continue

            geometry = task.geometry
            perfect_sorting = "".join(map(str, geometry.filling))
            parallel_decoder = geometry.parallel_decoder

            if parallel_decoder:
                cluster_indices = parallel_decoder.get_cluster_indices()
            else:
                cluster_indices = {(0, 0): list(range(len(perfect_sorting)))}

            shot_iter = filter(
                lambda shot: shot.shot_status == QuEraShotStatusCode.Completed,
                task.result().shot_outputs,
            )

            for shot, (cluster_coordinate, cluster_index) in product(
                shot_iter, cluster_indices.items()
            ):
                pre_sequence = "".join(
                    map(
                        str,
                        (shot.pre_sequence[index] for index in cluster_index),
                    )
                )

                post_sequence = np.asarray(
                    [shot.post_sequence[index] for index in cluster_index],
                    dtype=np.int8,
                )

                pfc_sorting = "".join(
                    [perfect_sorting[index] for index in cluster_index]
                )

                key = (
                    task_number,
                    cluster_coordinate,
                    pfc_sorting,
                    pre_sequence,
                )

                index.append(key)
                data.append(post_sequence)

            metas.append(task.metadata)
            geos.append(task.geometry)

        index = pd.MultiIndex.from_tuples(
            index, names=["task_number", "cluster", "perfect_sorting", "pre_sequence"]
        )

        df = pd.DataFrame(data, index=index)
        df.sort_index(axis="index")

        rept = None
        if self.name is None:
            rept = Report(df, metas, geos, "Remote")
        else:
            rept = Report(df, metas, geos, self.name)

        return rept


@RemoteBatch.set_serializer
def _serialize(obj: RemoteBatch) -> Dict[str, Any]:
    return {
        "source": None,
        "tasks": [(k, v) for k, v in obj.tasks.items()],
        "name": obj.name,
    }


@RemoteBatch.set_deserializer
def _deserialize(d: Dict[str, Any]) -> RemoteBatch:
    d["tasks"] = OrderedDict(d["tasks"])
    return RemoteBatch(**d)



from bloqade import start, var, cast, dumps, loads
from bloqade.atom_arrangement import Chain
import numpy as np
from beartype.typing import Dict
from scipy.stats import ks_2samp


def test_integration_1():
    batch = (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, "ramp_time", 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, "ramp_time", 0.1], [0, 10, 10, 0])
        .assign(ramp_time=3.0, r=8)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0, interaction_picture=True)
    )

    batch_str = dumps(batch)
    batch2 = loads(batch_str)
    assert isinstance(batch2, type(batch))
    batch2.report().bitstrings()


def test_integration_2():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, "ramp_time", 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.uniform.piecewise_constant(
            [0.1, ramp_time / 2, ramp_time / 2, 0.1], [0, 0, np.pi, np.pi]
        )
        .assign(ramp_time=3.0, r=6)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_3():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.uniform.piecewise_constant(
            [0.1, ramp_time / 2, ramp_time / 2, 0.1], [0, 0, np.pi, np.pi]
        )
        .amplitude.scale("rabi_mask")
        .fn(lambda t: 4 * np.sin(3 * t), ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1], r=6)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_4():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.scale("rabi_mask")
        .fn(lambda t: 4 * np.sin(3 * t), ramp_time + 0.2)
        .amplitude.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1], r=6)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_5():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, r=6)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_6():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .location(1)
        .constant(3.0, ramp_time + 0.2)
        .location(0)
        .linear(2.0, 0, ramp_time + 0.2)
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, r=6)
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_serialization():
    ramp_time = var("ramp_time")
    batch = (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.scale("rabi_mask")
        .piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1])
        .batch_assign(r=np.linspace(0.1, 4, 5).tolist())
        .bloqade.python()
        ._compile(100)
    )

    obj_str = dumps(batch)
    batch2 = loads(obj_str)
    assert isinstance(batch2, type(batch))


def KS_test(
    lhs_counts: Dict[str, int], rhs_counts: Dict[str, int], alpha: float = 0.05
) -> None:
    lhs_samples = []
    rhs_samples = []

    for bitstring, count in lhs_counts.items():
        lhs_samples += [int(bitstring, 2)] * count

    for bitstring, count in rhs_counts.items():
        rhs_samples += [int(bitstring, 2)] * count

    result = ks_2samp(lhs_samples, rhs_samples, method="exact")

    assert result.pvalue > alpha


def test_bloqade_against_braket():
    np.random.seed(9123892)
    durations = cast([0.1, 0.1, 0.1])

    prog = (
        Chain(3, lattice_spacing=6.1)
        .rydberg.detuning.uniform.piecewise_linear(durations, [-20, -20, "d", "d"])
        .amplitude.uniform.piecewise_linear(durations, [0, 15, 15, 0])
        .phase.uniform.constant(0.3, sum(durations))
        .batch_assign(d=[10, 20])
    )

    nshots = 1000
    a = prog.bloqade.python().run(nshots, cache_matrices=True).report().counts
    b = prog.braket.local_emulator().run(nshots).report().counts

    for lhs, rhs in zip(a, b):
        KS_test(lhs, rhs)


def test_bloqade_against_braket_2():
    np.random.seed(192839812)
    durations = cast([0.1, 0.1, 0.1])
    values = [0, 15, 15, 0]

    prog_1 = (
        Chain(3, lattice_spacing=6.1)
        .rydberg.detuning.uniform.piecewise_linear(durations, [-20, -20, "d", "d"])
        .amplitude.uniform.piecewise_linear(durations, values)
        .batch_assign(d=[10, 20])
    )
    prog_2 = (
        Chain(3, lattice_spacing=6.1)
        .rydberg.detuning.uniform.piecewise_linear(durations, [-20, -20, "d", "d"])
        .amplitude.location(0)
        .piecewise_linear(durations, values)
        .amplitude.location(1)
        .piecewise_linear(durations, values)
        .amplitude.location(2)
        .piecewise_linear(durations, values)
        .phase.location(0)
        .constant(0.0, sum(durations))
        .batch_assign(d=[10, 20])
    )

    nshots = 1000
    a = prog_2.bloqade.python().run(nshots, cache_matrices=True).report().counts
    b = prog_1.braket.local_emulator().run(nshots).report().counts

    for lhs, rhs in zip(a, b):
        KS_test(lhs, rhs)



from bloqade.builder.typing import ScalarType
from bloqade.builder.start import ProgramStart
from bloqade.ir.scalar import Scalar, Literal, cast
from bloqade.ir.location.transform import TransformTrait
from bloqade.ir.tree_print import Printer

from pydantic.dataclasses import dataclass
from beartype.typing import List, Tuple, Generator
from beartype import beartype
from enum import Enum
import plotext as pltxt
import sys
import numpy as np
from numpy.typing import NDArray
from bloqade.visualization import get_atom_arrangement_figure
from bloqade.visualization import display_ir


class SiteFilling(int, Enum):
    filled = 1
    vacant = 0

    def print_node(self) -> str:
        return self.name

    def children(self) -> List:
        return []


@dataclass(init=False)
class LocationInfo:
    position: Tuple[Scalar, Scalar]
    filling: SiteFilling

    @beartype
    def __init__(self, position: Tuple[ScalarType, ScalarType], filled: bool):
        if filled:
            self.filling = SiteFilling.filled
        else:
            self.filling = SiteFilling.vacant

        self.position = tuple(cast(ele) for ele in position)

    def print_node(self) -> str:
        return f"Location: {self.filling.name}"

    def children(self) -> List[Scalar]:
        return {"x": self.position[0], "y": self.position[1]}


@dataclass(init=False)
class AtomArrangement(ProgramStart, TransformTrait):
    def __str__(self) -> str:
        def is_literal(x):
            return isinstance(x, Literal)

        plot_results = all(
            is_literal(info.position[0]) and is_literal(info.position[1])
            for info in self.enumerate()
        )

        if plot_results:
            xs = [float(info.position[0].value) for info in self.enumerate()]
            ys = [float(info.position[1].value) for info in self.enumerate()]
            filling = [bool(info.filling.value) for info in self.enumerate()]
            xs_filled = [x for x, filled in zip(xs, filling) if filled]
            ys_filled = [y for y, filled in zip(ys, filling) if filled]
            xs_vacant = [x for x, filled in zip(xs, filling) if not filled]
            ys_vacant = [y for y, filled in zip(ys, filling) if not filled]

            pltxt.clear_figure()
            pltxt.limit_size(False, False)
            pltxt.plot_size(80, 24)
            pltxt.canvas_color("default")
            pltxt.axes_color("default")
            pltxt.ticks_color("white")
            pltxt.title("Atom Positions")
            pltxt.xlabel("x (um)")
            pltxt.ylabel("y (um)")

            marker = "."
            if sys.stdout.encoding.lower().startswith("utf"):
                marker = "dot"

            pltxt.scatter(xs_filled, ys_filled, color=(100, 55, 255), marker=marker)
            if len(xs_vacant) > 0:
                pltxt.scatter(
                    xs_vacant, ys_vacant, color="white", label="vacant", marker=marker
                )

            return pltxt.build()
        else:
            ph = Printer()
            ph.print(self)
            return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def print_node(self) -> str:
        return "AtomArrangement"

    def children(self) -> List[LocationInfo]:
        return list(self.enumerate())

    def enumerate(self) -> Generator[LocationInfo, None, None]:
        """enumerate all locations in the register."""
        raise NotImplementedError

    def figure(self, fig_kwargs=None, **assignments):
        """obtain a figure object from the atom arrangement."""
        return get_atom_arrangement_figure(self, fig_kwargs, **assignments)

    def show(self, **assignments) -> None:
        display_ir(self, assignments)

    def rydberg_interaction(self, **assignments) -> NDArray:
        """calculate the Rydberg interaction matrix.

        Args:
            **assignments: the values to assign to the variables in the register.

        Returns:
            NDArray: the Rydberg interaction matrix in the lower triangular form.

        """

        from bloqade.constants import RB_C6

        # calculate the Interaction matrix
        V_ij = np.zeros((self.n_sites, self.n_sites))
        for i, site_i in enumerate(self.enumerate()):
            pos_i = np.array([float(ele(**assignments)) for ele in site_i.position])

            for j, site_j in enumerate(self.enumerate()):
                if j >= i:
                    break  # enforce lower triangular form

                pos_j = np.array([float(ele(**assignments)) for ele in site_j.position])
                r_ij = np.linalg.norm(pos_i - pos_j)

                V_ij[i, j] = RB_C6 / r_ij**6

        return V_ij

    @property
    def n_atoms(self) -> int:
        """number of atoms (filled sites) in the register."""
        raise NotImplementedError

    @property
    def n_sites(self) -> int:
        """number of sites in the register."""
        raise NotImplementedError

    @property
    def n_vacant(self) -> int:
        """number of vacant sites in the register."""
        raise NotImplementedError

    @property
    def n_dims(self) -> int:
        """number of dimensions in the register."""
        raise NotImplementedError


@dataclass(init=False)
class ParallelRegister(ProgramStart):
    """Parallel Register"""

    __match_args__ = ("_register", "_cluster_spacing")

    register_locations: List[List[Scalar]]
    register_filling: List[int]
    shift_vectors: List[List[Scalar]]

    @beartype
    def __init__(self, register: AtomArrangement, cluster_spacing: ScalarType):
        self._register = register
        self._cluster_spacing = cast(cluster_spacing)

        if register.n_atoms > 0:
            # calculate bounding box
            # of this register
            location_iter = register.enumerate()
            (x, y) = next(location_iter).position
            x_min = x
            x_max = x
            y_min = y
            y_max = y

            for location_info in location_iter:
                (x, y) = location_info.position
                x_min = x.min(x_min)
                x_max = x.max(x_max)
                y_min = y.min(y_min)
                y_max = y.max(y_max)

            shift_x = (x_max - x_min) + cluster_spacing
            shift_y = (y_max - y_min) + cluster_spacing

            register_locations = [
                list(location_info.position) for location_info in register.enumerate()
            ]
            register_filling = [
                location_info.filling.value for location_info in register.enumerate()
            ]
            shift_vectors = [[shift_x, cast(0)], [cast(0), shift_y]]
        else:
            raise ValueError("No locations to parallelize.")

        self.register_locations = register_locations
        self.register_filling = register_filling
        self.shift_vectors = shift_vectors
        super().__init__(self)



import simplejson as json
import logging
import uuid
from typing import Optional, Union, Dict, Tuple
import requests
from requests_sigv4 import Sigv4Request
import time


class ApiRequest:
    """Class that defines base methods for API requests."""

    class ApiRequestError(Exception):
        pass

    class InvalidResponseError(ApiRequestError):
        pass

    def __init__(
        self,
        api_hostname: str,
        qpu_id: str,
        api_stage="v0",
        virtual_queue: Optional[str] = None,
        proxy: Optional[str] = None,
    ):
        """
        Create an instance of `ApiRequest`.
        @param api_hostname: hostname of the API instance.
        @param qpu_id: The QPU ID, for example `qpu1-mock`.
        @param api_stage: Specify which version of the API to call from this object.
        @param virtual_queue: Optional, the virtual queue to use for the API.
        @param proxy: Optional, the hostname for running the API via some proxy
        endpoint.
        """

        if proxy is None:
            self.hostname = None
            self.aws_host = api_hostname
            uri_with_version = api_hostname + f"/{api_stage}"
        else:
            self.hostname = api_hostname
            self.aws_host = proxy
            uri_with_version = proxy + f"/{api_stage}"

        self.base_url = "https://" + uri_with_version
        self.qpu_id = qpu_id
        self.virtual_queue = virtual_queue
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _result_as_json(result: requests.Response, **options) -> Dict:
        content_type = result.headers["Content-Type"]
        if content_type != "application/json":
            raise ApiRequest.InvalidResponseError(
                f"Expected Content-Type application/json, but {content_type} found."
            )

        if len(result.content) == 0:
            return {}

        return json.loads(result.content, **options)

    def _generate_headers(self, base: Optional[dict] = None) -> Dict:
        if base is None:
            header = {"Content-Type": "application/json"}
        else:
            header = dict(base)
            header["Content-Type"] = "application/json"

        if self.hostname is not None:
            header["Host"] = self.hostname

        if self.virtual_queue is not None:
            header["virtual_queue"] = self.virtual_queue

        return header

    def _get_path(self, *path_list: str):
        return "/".join((self.base_url, self.qpu_id) + path_list)

    def _prepare_request(self, *path_list: str) -> Tuple[str, Dict[str, str]]:
        headers = self._generate_headers()
        url = self._get_path(*path_list)
        return url, headers

    def _post(
        self, url: str, headers: Dict[str, str], content: str
    ) -> requests.Response:
        raise NotImplementedError

    def _get(self, url: str, headers: Dict[str, str]) -> requests.Response:
        raise NotImplementedError

    def _put(self, url: str, headers: Dict[str, str]) -> requests.Response:
        raise NotImplementedError

    # public API
    def post(self, *path_list: str, content: Dict[str, str] = {}) -> requests.Response:
        url, headers = self._prepare_request(*path_list)

        if isinstance(content, dict):
            content = json.dumps(content)

        self.logger.info(f'POSTing to "{url}".')
        response = self._post(url, headers, content)
        self.logger.debug(f"API return status {response.status_code}.")
        return response

    def get(self, *path_list: str) -> requests.Response:
        url, headers = self._prepare_request(*path_list)
        self.logger.info(f'GETting "{url}".')
        response = self._get(url, headers)
        self.logger.debug(f"API return status {response.status_code}.")
        return response

    def put(self, *path_list: str) -> requests.Response:
        url, headers = self._prepare_request(*path_list)
        self.logger.info(f'PUTting "{url}".')
        response = self._put(url, headers)
        self.logger.debug(f"API return status {response.status_code}.")
        return response


class AwsApiRequest(ApiRequest):
    def __init__(
        self,
        api_hostname: str,
        qpu_id: str,
        api_stage="v0",
        virtual_queue: Optional[str] = None,
        proxy: Optional[str] = None,
        # Sigv4Request arguments
        region: str = "us-east-1",
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        session_token: Optional[str] = None,
        session_expires: int = 3600,
        role_arn: Optional[str] = None,
        role_session_name: str = "awsrequest",
        profile: Optional[str] = None,
    ):
        """
        Create an instance of `AwsApiRequest`.
        @param api_hostname: hostname of the API instance.
        @param qpu_id: The QPU ID, for example `qpu1-mock`.
        @param api_stage: Specify which version of the API to call from this object.
        @param virtual_queue: Optional, the virtual queue to use for the API.
        @param proxy: Optional, the hostname for running the API via some proxy
        endpoint.
        @param region: AWS region, default value: "us-east-1"
        @param access_key: Optional, AWS account access key
        @param secret_key: Optional, AWS account secret key
        @param session_token: Optional, AWS session token
        @param session_expires: int, time before current tokens expire, default value
        3600
        @param role_arn: Optional, AWS role ARN
        @param role_session_name: AWS role session name, defualy value: 'awsrequest',
        @param profile: Optional, AWS profile to use credentials for.
        """
        super().__init__(
            api_hostname,
            qpu_id,
            virtual_queue=virtual_queue,
            api_stage=api_stage,
            proxy=proxy,
        )

        self.request = Sigv4Request(
            region=region,
            access_key=access_key,
            secret_key=secret_key,
            session_token=session_token,
            session_expires=session_expires,
            role_arn=role_arn,
            role_session_name=role_session_name,
            profile=profile,
        )

    def _post(
        self, url: str, headers: Dict[str, str], content: Dict[str, str]
    ) -> requests.Response:
        return self.request.post(url, headers=headers, data=content)

    def _get(self, url: str, headers: Dict[str, str]) -> requests.Response:
        return self.request.get(url, headers=headers)

    def _put(self, url: str, headers: Dict[str, str]) -> requests.Response:
        return self.request.put(url, headers=headers)


class QueueApi:
    """Simple interface to the QCS task API.

    Example (replace URIs, QPU ID with correct values):
    >>> task_json = {
    ...     "nshots": 10,
    ...     "lattice": {
    ...         "sites":[[0,0]],
    ...         "filling":[1],
    ...     },
    ...     "effective_hamiltonian": {
    ...         "rydberg": {
    ...             "rabi_frequency_amplitude":{
    ...                 "global": {
    ...                     "times":[0.0, 0.1e-6, 3.9e-6, 4.0e-6],
    ...                     "values":[0.0, 15.0e6, 15.0e6, 0.0],
    ...                 }
    ...             },
    ...             "rabi_frequency_phase": {
    ...                 "global": {
    ...                 "times":[0.0, 4.0e-6],
    ...                 "values":[0.0, 0.0],
    ...                 }
    ...             },
    ...             "detuning":{
    ...                 "global": {
    ...                     "times":[0.0, 4.0e-6],
    ...                     "values":[0.0, 0.0],
    ...                 }
    ...             }
    ...         }
    ...     }
    ... }
    To Use this class with API-Gateway:
    >>> api_hostname = "XXX.execute-api.us-east-1.amazonaws.com"
    >>> vpce_uri = "vpce-XXX-XXX.execute-api.us-east-1.vpce.amazonaws.com"
    >>> api = QueueApi(api_hostname, "qpu1-mock", proxy=vpce_uri)
    >>> print(api.get_capabilities())
    """

    bad_request = (
        "The request is invalid. This may indicate an error when parsing a parameter."
    )
    bad_submit_task_request = (
        "The request is invalid. This may indicate an error when parsing a parameter, "
        "or an error when parsing or validating the request body. The response body "
        "may contain a JSON array with a list of validation errors."
    )

    qpu_not_found = (
        "The QPU targeted by the request does not exist, or the user "
        "making the request is not authorized to access the targeted QPU."
    )
    qpu_or_task_not_found = (
        "The QPU or task targeted by the request does not exist, "
        "or the user making the request is not authorized to access the targeted QPU "
        "or task."
    )

    class QueueApiError(RuntimeError):
        pass

    class QpuBusyError(QueueApiError):
        pass

    class NotFound(QueueApiError):
        pass

    class InvalidRequestError(QueueApiError):
        pass

    class InvalidResponseError(QueueApiError):
        pass

    class ValidationError(QueueApiError):
        pass

    class AuthenticationError(QueueApiError):
        pass

    def __init__(
        self,
        api_hostname: str,
        qpu_id: str,
        api_stage="v0",
        virtual_queue: Optional[str] = None,
        proxy: Optional[str] = None,
        **request_sigv4_kwargs,
    ):
        """
        Create an instance of `QueueApi`.
        @param api_hostname: hostname of the API instance.
        @param qpu_id: The QPU ID, for example `qpu1-mock`.
        @param api_stage: Specify which version of the API to call from this object.
        @param virtual_queue: Optional, the virtual queue to use for the API.
        @param proxy: Optional, the hostname for running the API via some proxy
        endpoint.

        request_sigv4_kwargs:

        @param region: AWS region, default value: "us-east-1"
        @param access_key: Optional, AWS account access key
        @param secret_key: Optional, AWS account secret key
        @param session_token: Optional, AWS session token
        @param session_expires: int, time before current tokens expire, default value
        3600
        @param role_arn: Optional, AWS role ARN
        @param role_session_name: AWS role session name, defualy value: 'awsrequest',
        @param profile: Optional, AWS profile to use credentials for.
        """
        self.api_http_request = AwsApiRequest(
            api_hostname,
            qpu_id,
            api_stage=api_stage,
            virtual_queue=virtual_queue,
            proxy=proxy,
            **request_sigv4_kwargs,
        )

        self.logger = logging.getLogger(self.__class__.__name__)

    def post_task(self, task_json: Union[str, dict]) -> str:
        """
        Submit a task to the QPU via the task API.
        @param content: Task specification as a JSON string or dictionary.
        """
        result = self.api_http_request.post("queue", "task", content=task_json)

        if result.status_code == 201:
            message = "Successfully posted task."
            self.logger.warning(message)
        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)
        elif result.status_code == 400:
            message = (
                "The request is invalid. This may indicate an error when parsing "
                "a parameter."
            )
            self.logger.error(message)
            raise QueueApi.InvalidRequestError(message)
        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)
        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

        result_json = ApiRequest._result_as_json(result)

        try:
            task_id = result_json["task_id"]
            self.logger.info(
                f"QPU {self.api_http_request.qpu_id} accepted "
                f"task with task id {task_id}."
            )
            return task_id
        except KeyError:
            raise QueueApi.InvalidResponseError('Response did not contain "task_id".')

    def get_capabilities(self) -> Dict:
        """
        Request the QPU capabilities from the task API.
        @return: dictionary containing different fields for capabilities.
        """
        result = self.api_http_request.get("capabilities")

        if result.status_code == 200:
            message = "Successfully fetched capabilities."
            self.logger.info(message)
        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)
        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)
        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

        return ApiRequest._result_as_json(result, use_decimal=True)

    def validate_task(self, task_json: Union[str, dict]) -> None:
        result = self.api_http_request.post("task", "validate", content=task_json)

        if result.status_code == 200:
            message = "Task passed validation."
            self.logger.info(message)
        elif result.status_code == 400:
            message = f"Task did not pass validation: {result.text}"
            self.logger.error(message)
            raise QueueApi.ValidationError(message)
        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)
        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)
        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

    def get_task_results(self, task_id: Union[str, uuid.UUID]) -> Dict:
        """
        Return task results as given by API.
        @return: Parsed JSON of the task results.
        @param task_id: Task ID.
        """

        queue_status = self.get_task_status_in_queue(task_id)
        # TODO: Revisit this mapping when the queue API is
        #       has task result fetching build into the API

        if queue_status in ("Created", "Enqueued", "Accepted"):
            return {"task_status": "Created", "shot_outputs": []}
        elif queue_status == "Executing":
            return {"task_status": "Running", "shot_outputs": []}
        elif queue_status in ("Failed", "Cancelled"):
            return {"task_status": queue_status, "shot_outputs": []}
        elif queue_status == "Unaccepted":
            raise QueueApi.ValidationError(
                f"Task: {task_id} has validation error, "
                "unable to fetch error message."
            )
        elif queue_status in ("Completed", "Partial"):
            pass
        else:
            raise QueueApi.QueueApiError(f"Undocumented queue status: {queue_status}")

        result = self.api_http_request.get("task", str(task_id), "results")

        if result.status_code == 200:
            message = f"Successfully fetch task results for task_id {task_id}"
            self.logger.debug(message)
        elif result.status_code == 400:
            message = QueueApi.bad_request
            self.logger.debug(message)
            raise QueueApi.InvalidRequestError(message)
        elif result.status_code == 404:
            message = QueueApi.qpu_or_task_not_found
            self.logger.debug(message)
            raise QueueApi.NotFound(message)

        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)
        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

        return ApiRequest._result_as_json(result)

    def poll_task_results(
        self, task_id: Union[str, uuid.UUID], polling_interval=3
    ) -> Dict:
        status = self.get_task_status_in_queue(task_id)
        while status not in ("Partial", "Completed", "Failed", "Cancelled"):
            status = self.get_task_status_in_queue(task_id)
            time.sleep(polling_interval)

        return self.get_task_results(task_id)

    def get_task_status_in_queue(self, task_id: Union[str, uuid.UUID]) -> str:
        """
        Request task status in a queue for a given task.
        @param task_id: Task ID.
        @return: Parsed JSON of the task status.
        """
        result = self.api_http_request.get("queue", "task", str(task_id))

        if result.status_code == 200:
            message = "Successfully checked queue."
            self.logger.debug(message)
        elif result.status_code == 400:
            message = (
                "The request is invalid. This may indicate an error when parsing a "
                "parameter."
            )
            self.logger.error(message)
            raise QueueApi.InvalidRequestError(message)
        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} or task {task_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)
        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

        result_json = ApiRequest._result_as_json(result)

        return result_json["status"]

    def cancel_task_in_queue(self, task_id: Union[str, uuid.UUID]):
        result = self.api_http_request.put("queue", "task", str(task_id), "cancel")

        if result.status_code == 200:
            message = "Task successfully cancelled"
            self.logger.debug(message)

        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)

        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} or task {task_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)

        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

    def get_task_summary(self, task_id: Union[str, uuid.UUID]) -> Dict:
        """
        Request the task summary for a given task. The summary contains the status of
        the current task.
        @param task_id: Task ID.
        @return: Parsed JSON of the task summary.
        @see: `TaskSummary` in https://github.com/QuEra-QCS/QCS-API/blob/master/qcs-api/openapi.yaml
        """

        if self.get_task_status_in_queue(task_id) != "Completed":
            raise QueueApi.QueueApiError(
                "Task is not completed, can't fetch task summary."
            )

        result = self.api_http_request.get("task", str(task_id))

        if result.status_code == 200:
            message = "Successfully checked task summary."
            self.logger.warning(message)

        elif result.status_code == 404:
            message = f"QPU {self.api_http_request.qpu_id} or task {task_id} not found."
            self.logger.error(message)
            raise QueueApi.NotFound(message)

        elif result.status_code == 403:
            message = "QPU return authentication error, check AWS credentials."
            self.logger.error(message)
            raise QueueApi.AuthenticationError(message)

        else:
            message = f"QPU returned unhandled status {result.status_code}."
            self.logger.error(message)
            raise QueueApi.QueueApiError(message)

        return ApiRequest._result_as_json(result)

    def is_task_stopped(self, task_id: Union[str, uuid.UUID]) -> bool:
        """
        Check whether a task is stopped (because it is completed, failed, or cancelled).
        @param task_id:
        @return: `True` if task is stopped.
        """
        return self.get_task_status_in_queue(task_id) in (
            "Completed",
            "Failed",
            "Cancelled",
        )



import simplejson as json
from typing import Any
from beartype.typing import Type, Callable, Dict, Union, TextIO
from beartype import beartype


__bloqade_package_loaded__ = False


def load_bloqade():
    import pkgutil
    import os

    # call this function to load all modules in this package
    # required because if no other modules are imported, the
    # Various classes will not be registered with Serializer
    # and the serialization will fail. only need to call this
    # function once per process hence the flag to prevent
    # multiple calls

    if not __bloqade_package_loaded__:
        path = os.path.abspath(os.path.join(os.path.dirname(__file__)))

        for loader, module_name, _ in pkgutil.walk_packages([path]):
            _module = loader.find_module(module_name).load_module(module_name)
            globals()[module_name] = _module

        globals()["__bloqade_package_loaded__"] = True


class Serializer(json.JSONEncoder):
    types = ()
    type_to_str = {}
    str_to_type = {}
    serializers = {}
    deserializers = {}

    @staticmethod
    @beartype
    def register(cls: Type):
        @beartype
        def _deserializer(d: Dict[str, Any]) -> cls:
            return cls(**d)

        @beartype
        def _serializer(obj: cls) -> Dict[str, Any]:
            return obj.__dict__

        @beartype
        def set_serializer(f: Callable):
            # TODO: check function signature
            setattr(cls, "__bloqade_serializer__", staticmethod(f))
            Serializer.serializers[cls] = cls.__bloqade_serializer__

        @beartype
        def set_deserializer(f: Callable):
            # TODO: check function signature
            setattr(cls, "__bloqade_deserializer__", staticmethod(f))
            Serializer.deserializers[cls] = cls.__bloqade_deserializer__

        type_name = f"{cls.__module__}.{cls.__name__}"
        Serializer.type_to_str[cls] = type_name
        Serializer.str_to_type[type_name] = cls
        Serializer.types += (cls,)
        setattr(cls, "set_serializer", staticmethod(beartype(set_serializer)))
        setattr(cls, "set_deserializer", staticmethod(beartype(set_deserializer)))
        cls.set_deserializer(_deserializer)
        cls.set_serializer(_serializer)

        return cls

    @classmethod
    def object_hook(cls, d: Any) -> Any:
        if isinstance(d, dict) and len(d) == 1:
            ((key, value),) = d.items()
            if key in cls.str_to_type:
                obj_cls = cls.str_to_type[key]
                deserialize = cls.deserializers.get(obj_cls)

                return deserialize(value)

        return d

    def default(self, o: Any) -> Any:
        if type(o) in self.serializers:
            cls = type(o)
            serializer = self.serializers.get(cls)

            return {self.type_to_str[cls]: serializer(o)}

        return super().default(o)


@beartype
def loads(s: str, use_decimal: bool = True, **json_kwargs):
    """Load object from string

    Args:
        s (str): the string to load
        use_decimal (bool, optional): use decimal.Decimal for numbers. Defaults to True.
        **json_kwargs: other arguments passed to json.loads

    Returns:
        Any: the deserialized object
    """
    load_bloqade()
    return json.loads(
        s, object_hook=Serializer.object_hook, use_decimal=use_decimal, **json_kwargs
    )


@beartype
def load(fp: Union[TextIO, str], use_decimal: bool = True, **json_kwargs):
    """Load object from file

    Args:
        fp (Union[TextIO, str]): the file path or file object
        use_decimal (bool, optional): use decimal.Decimal for numbers. Defaults to True.
        **json_kwargs: other arguments passed to json.load

    Returns:
        Any: the deserialized object
    """
    load_bloqade()
    if isinstance(fp, str):
        with open(fp, "r") as f:
            return json.load(
                f,
                object_hook=Serializer.object_hook,
                use_decimal=use_decimal,
                **json_kwargs,
            )
    else:
        return json.load(
            fp,
            object_hook=Serializer.object_hook,
            use_decimal=use_decimal,
            **json_kwargs,
        )


@beartype
def dumps(
    o: Any,
    use_decimal: bool = True,
    **json_kwargs,
) -> str:
    """Serialize object to string

    Args:
        o (Any): the object to serialize
        use_decimal (bool, optional): use decimal.Decimal for numbers. Defaults to True.
        **json_kwargs: other arguments passed to json.dumps

    Returns:
        str: the serialized object as a string
    """
    if not isinstance(o, Serializer.types):
        raise TypeError(
            f"Object of type {type(o)} is not JSON serializable. "
            f"Only {Serializer.types} are supported."
        )
    return json.dumps(o, cls=Serializer, use_decimal=use_decimal, **json_kwargs)


@beartype
def save(
    o: Any,
    fp: Union[TextIO, str],
    use_decimal=True,
    **json_kwargs,
) -> None:
    """Serialize object to file

    Args:
        o (Any): the object to serialize
        fp (Union[TextIO, str]): the file path or file object
        use_decimal (bool, optional): use decimal.Decimal for numbers. Defaults to True.
        **json_kwargs: other arguments passed to json.dump

    Returns:
        None
    """
    if not isinstance(o, Serializer.types):
        raise TypeError(
            f"Object of type {type(o)} is not JSON serializable. "
            f"Only {Serializer.types} are supported."
        )
    if isinstance(fp, str):
        with open(fp, "w") as f:
            json.dump(o, f, cls=Serializer, use_decimal=use_decimal, **json_kwargs)
    else:
        json.dump(o, fp, cls=Serializer, use_decimal=use_decimal, **json_kwargs)



from bloqade.ir.location.bravais import (
    Lieb,
    Square,
    Rectangular,
    Honeycomb,
    Kagome,
    Triangular,
    Chain,
)
from bloqade.ir.location.bravais import Cell
from bloqade import cast
from math import sqrt
from bloqade.codegen.common.json import BloqadeIRSerializer


def test_square():
    lattice = Square(3, lattice_spacing=2.0)

    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast([(0, 0), (0, 2), (0, 4), (2, 0), (2, 2), (2, 4), (4, 0), (4, 2), (4, 4)])
    )

    assert positions == positions_expected


def test_rectangular():
    lattice = Rectangular(2, 3, lattice_spacing_x=0.5, lattice_spacing_y=2)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast([(0, 0), (0, 2.0), (0, 4.0), (0.5, 0), (0.5, 2.0), (0.5, 4.0)])
    )
    assert positions == positions_expected


def test_rectagnular_default_yscale():
    lattice = Rectangular(2, 3, lattice_spacing_x=0.5)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast([(0, 0), (0, 1.0), (0, 2.0), (0.5, 0), (0.5, 1.0), (0.5, 2.0)])
    )
    assert positions == positions_expected


def test_kagome():
    lattice = Kagome(2, lattice_spacing=2.0)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (1.0, 0),
                (2, 0),
                (3, 0),
                (0.5, sqrt(3) * 0.5),
                (2.5, sqrt(3) * 0.5),
                (1, sqrt(3)),
                (3, sqrt(3)),
                (2, sqrt(3)),
                (4, sqrt(3)),
                (1.5, sqrt(3) * 1.5),
                (3.5, sqrt(3) * 1.5),
            ]
        )
    )

    assert positions == positions_expected


def test_triangular():
    lattice = Triangular(2, lattice_spacing=2.0)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(cast([(0, 0), (2, 0), (1.0, sqrt(3)), (3, sqrt(3))]))

    assert positions == positions_expected


def test_honeycomb():
    lattice = Honeycomb(2, lattice_spacing=2)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0.0, 0),
                (2.0, 0),
                (1.0, 1 / sqrt(3)),
                (3.0, 1 / sqrt(3)),
                (1.0, sqrt(3)),
                (3.0, sqrt(3)),
                (2.0, sqrt(3) + 1 / sqrt(3)),
                (4.0, sqrt(3) + 1 / sqrt(3)),
            ]
        )
    )

    assert positions == positions_expected
    assert lattice.n_dims == 2


def test_lieb():
    lattice = Lieb(2, lattice_spacing=2)
    positions = set(map(lambda info: info.position, lattice.enumerate()))
    positions_expected = set(
        cast(
            [
                (0, 0),
                (2, 0),
                (0, 2),
                (2, 2),
                (1, 0),
                (3, 0),
                (1, 2),
                (3, 2),
                (0, 1),
                (2, 1),
                (0, 3),
                (2, 3),
            ]
        )
    )

    assert positions == positions_expected
    assert lattice.n_dims == 2


def test_scale_lattice():
    lattice = Triangular(2, lattice_spacing=1)
    latt2 = lattice.scale(2)
    positions = set(map(lambda info: info.position, latt2.enumerate()))
    positions_expected = set(cast([(0, 0), (2, 0), (1.0, sqrt(3)), (3, sqrt(3))]))

    assert positions == positions_expected
    assert lattice.n_dims == 2


def test_chain():
    lattice = Chain(4, lattice_spacing=1, vertical_chain=False)
    latt2 = lattice.scale(2)
    positions = set(map(lambda info: info.position, latt2.enumerate()))
    positions_expected = set(cast([(0, 0), (2, 0), (4, 0), (6, 0)]))

    assert positions == positions_expected
    assert lattice.n_dims == 1

    vertical_lattice = Chain(4, lattice_spacing=1, vertical_chain=True)
    positions = set(map(lambda info: info.position, vertical_lattice.enumerate()))
    positions_expected = set(cast([(0, 0), (0, 1), (0, 2), (0, 3)]))

    assert positions == positions_expected
    assert vertical_lattice.n_dims == 1

    lattice_dict = BloqadeIRSerializer().default(vertical_lattice)

    assert lattice_dict["chain"]["L"] == 4
    assert lattice_dict["chain"]["vertical_chain"] is True
    assert lattice_dict["chain"]["lattice_spacing"] == {"literal": {"value": "1"}}

    assert lattice_dict


def test_cell():
    c = Cell(10, 2)

    assert c.natoms == 10
    assert c.ndims == 2



from bloqade.submission.base import SubmissionBackend
from bloqade.submission.ir.braket import (
    from_braket_task_results,
    from_braket_status_codes,
    to_braket_task,
)
from bloqade.submission.ir.task_results import (
    QuEraTaskStatusCode,
    QuEraTaskResults,
)
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from braket.aws import AwsDevice, AwsQuantumTask
from beartype.typing import Optional
from pydantic import PrivateAttr
import bloqade


class BraketBackend(SubmissionBackend):
    device_arn: str = "arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    _device: Optional[AwsDevice] = PrivateAttr(default=None)

    @property
    def device(self) -> AwsDevice:
        if self._device is None:
            self._device = AwsDevice(self.device_arn)
            user_agent = f"Bloqade/{bloqade.__version__}"
            self._device.aws_session.add_braket_user_agent(user_agent)

        return self._device

    def submit_task(self, task_ir: QuEraTaskSpecification) -> str:
        shots, ahs_program = to_braket_task(task_ir)
        task = self.device.run(ahs_program, shots=shots)
        return task.id

    def task_results(self, task_id: str) -> QuEraTaskResults:
        return from_braket_task_results(AwsQuantumTask(task_id).result())

    def cancel_task(self, task_id: str) -> None:
        AwsQuantumTask(task_id).cancel()

    def task_status(self, task_id: str) -> QuEraTaskStatusCode:
        return from_braket_status_codes(AwsQuantumTask(task_id).state())

    def validate_task(self, task_ir: QuEraTaskSpecification):
        pass

    # def validate_task(self, task_ir: QuEraTaskSpecification):
    #     try:
    #         task_id = self.submit_task(task_ir)
    #     except Exception as e:
    #         if "ValidationException" in str(e) and "validation error" in str(e):
    #             raise ValidationError(str(e))
    #         else:
    #             raise e

    #     # don't want the task to actually run
    #     try:
    #         self.cancel_task(task_id)
    #     except Exception as e:
    #         return



from bloqade.emulate.sparse_operator import (
    SparseMatrixCSC,
    SparseMatrixCSR,
    IndexMapping,
    _csc_matvec_impl,
    _csr_matvec_impl,
    _index_mapping_col_sliced,
    _index_mapping_row_sliced,
    _index_mapping_impl,
    _index_mapping_row_col_sliced,
)
from scipy.sparse import random
import numpy as np

np.random.seed(12039)


def test_csr():
    A = random(10, 10, density=0.5, format="csr")
    B = SparseMatrixCSR.create(A)
    assert np.allclose(A.toarray(), B.toarray())

    v = np.random.normal(size=10)
    a = 3

    expected_result = a * A.dot(v) + v
    result = v.copy()
    B.matvec(v, out=result, scale=a)

    assert np.allclose(result, expected_result)

    result = v.copy()

    _csr_matvec_impl.py_func(B.shape[1], B.data, B.indices, B.indptr, a, v, result)

    assert np.allclose(result, expected_result)


def test_csc():
    A = random(10, 10, density=0.5, format="csc")
    B = SparseMatrixCSC.create(A)
    assert np.allclose(A.toarray(), B.toarray())

    v = np.random.normal(size=10)
    a = 3

    expected_result = a * A.dot(v) + v
    result = v.copy()
    B.matvec(v, out=result, scale=a)

    assert np.allclose(result, expected_result)

    result = v.copy()

    _csc_matvec_impl.py_func(B.shape[1], B.data, B.indices, B.indptr, a, v, result)

    assert np.allclose(result, expected_result)


def test_index_mapping():
    B = IndexMapping(10, slice(0, None, 2), slice(1, None, 2))
    A = B.tocsr().toarray()

    v = np.random.normal(size=10)
    a = 3

    expected_result = a * A.dot(v) + v

    result = v.copy()
    B.matvec(v, out=result, scale=a)

    print(result)
    print(expected_result)
    print("---------------")

    assert np.allclose(result, expected_result)

    result = v.copy()

    _index_mapping_row_col_sliced.py_func(
        a,
        v[B.col_indices],
        result[B.row_indices],
    )

    print(result)
    print(expected_result)

    assert np.allclose(result, expected_result)


def test_index_mapping_2():
    input = np.random.permutation(np.arange(10))

    B = IndexMapping(10, input, slice(None))
    A = B.tocsr().toarray()

    v = np.random.normal(size=10)
    a = 3
    print(A)

    expected_result = a * A.dot(v) + v

    result = v.copy()
    B.matvec(v, out=result, scale=a)

    print(result)
    print(expected_result)
    print("---------------")

    assert np.allclose(result, expected_result)

    result = v.copy()

    _index_mapping_col_sliced.py_func(
        B.row_indices,
        a,
        v[B.col_indices],
        result,
    )

    print(result)
    print(expected_result)

    assert np.allclose(result, expected_result)

    B = B.T
    A = B.tocsr().toarray()

    expected_result = a * A.dot(v) + v

    result = v.copy()
    B.matvec(v, out=result, scale=a)

    print(result)
    print(expected_result)
    print("---------------")

    assert np.allclose(result, expected_result)

    result = v.copy()

    _index_mapping_row_sliced.py_func(
        B.col_indices,
        a,
        v,
        result[B.row_indices],
    )

    print(result)
    print(expected_result)

    assert np.allclose(result, expected_result)


def test_index_mapping_3():
    input = np.random.permutation(np.arange(10))
    output = np.random.permutation(np.arange(10))

    B = IndexMapping(10, input, output)
    A = B.tocsr().toarray()

    v = np.random.normal(size=10)
    a = 3

    expected_result = a * A.dot(v) + v

    result = v.copy()
    B.matvec(v, out=result, scale=a)

    print(result)
    print(expected_result)
    print("---------------")

    assert np.allclose(result, expected_result)

    result = v.copy()

    _index_mapping_impl.py_func(
        B.col_indices,
        B.row_indices,
        a,
        v,
        result,
    )

    print(result)
    print(expected_result)

    assert np.allclose(result, expected_result)



# from bloqade.julia.types import Vector, Int64

# print(Vector[Int64]([1, 2, 3]))


class Vector:
    def __init__(self, *params) -> None:
        self.expr = "Vector"
        self.params = params

    def __getitem__(self, *params) -> "Vector":
        return Vector(*params)


Vector(int)



from bloqade import start
from bloqade.ir.location import Square
from bloqade.ir import Constant, Linear
import json
import numpy as np
import pytest


def f(x):
    return float(x)


def fvec(v):
    return list(map(f, v))


def test_location_exceedloc():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
            2 * Constant(0.5, duration=0.5)
        )
        start.rydberg.detuning.location(0).apply(seq).quera.mock()._compile(10)


def test_hyperfine_schema():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
            2 * Constant(0.5, duration=0.5)
        )
        Square(1).hyperfine.detuning.location(0).apply(seq).quera.mock()._compile(
            shots=10
        )


def test_local_no_global():
    seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
        2 * Constant(0.5, duration=0.5)
    )
    batch = (
        Square(1)
        .rydberg.detuning.location(0, 0.5)
        .apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(batch.json())

    print(panel)
    panel = panel["bloqade.task.batch.RemoteBatch"]

    task = panel["tasks"][0][1]["bloqade.task.quera.QuEraTask"]
    ir = task["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]

    assert all(fvec(detune_ir["local"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["local"]["values"]) == np.array([0, 1, 1]) * 1e6)
    assert detune_ir["local"]["lattice_site_coefficients"] == [0.5]

    # should have global even without assign:
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 0]))


def test_local_global():
    seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
        2 * Constant(0.5, duration=0.5)
    )
    seq_global = Constant(0.55, duration=1.0)

    job = (
        Square(1)
        .rydberg.detuning.uniform.apply(seq_global)
        .location(0, 0.5)
        .apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())
    panel = panel["bloqade.task.batch.RemoteBatch"]

    print(panel)
    task = panel["tasks"][0][1]["bloqade.task.quera.QuEraTask"]
    ir = task["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["local"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["local"]["values"]) == np.array([0, 1, 1]) * 1e6)
    assert fvec(detune_ir["local"]["lattice_site_coefficients"]) == [0.5]

    # global :
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0.55, 0.55]) * 1e6)



# import pytest
import pytest
from bloqade import start
from bloqade.serialize import loads, dumps

# import numpy as np


# @pytest.mark.skip(reason="not a good unit test, skip for now")
def test_braket_simulator_getbitstring():
    program = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=[0.05, 1, 0.05], values=[0.0, 15.8, 15.8, 0.0]
        )
        .detuning.uniform.piecewise_linear(durations=[1.1], values=[0.0, 0.0])
    )

    output = program.braket.local_emulator().run(shots=10).report()

    assert len(output.bitstrings()[0].flatten()) == 10


def test_serialize_braket_simulator_batch():
    program = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=[0.05, 1, 0.05], values=[0.0, 15.8, 15.8, 0.0]
        )
        .detuning.uniform.piecewise_linear(durations=[1.1], values=[0.0, 0.0])
    )

    output = program.braket.local_emulator().run(shots=10)

    output_str = dumps(output)
    assert isinstance(loads(output_str), type(output))
    assert loads(output_str).tasks == output.tasks
    assert loads(output_str).name == output.name


def test_error():
    program = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=[0.05, 1, 0.05], values=[0.0, 15.8, 15.8, 0.0]
        )
        .detuning.uniform.piecewise_linear(durations=[1.1], values=[0.0, 0.0])
    )

    output = program.braket.local_emulator()._compile(10)

    with pytest.raises(ValueError):
        output.tasks[0].result()

    output_str = dumps(output)
    assert isinstance(loads(output_str), type(output))
    assert loads(output_str).tasks == output.tasks
    assert loads(output_str).name == output.name


"""
from bloqade.ir.location import Square
if __name__ == "__main__":
    simulator_job = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            [0.1, "run_time", 0.1], [0, 15, 15, 0]
        )
        .batch_assign(run_time=np.linspace(0.1, 3.8, 20))
        .braket_local_simulator(1000)
    )

    rydberg_densities = simulator_job.run_async().report().rydberg_densities()

    rydberg_densities = (
        simulator_job.run_async(multiprocessing=True).report().rydberg_densities()
    )

    # durations for rabi and detuning
    durations = [0.3, 1.6, 0.3]

    mis_udg_program = (
        Square(4, 5.5)
        .apply_defect_density(0.5)
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations, [0.0, 15.0, 15.0, 0.0]
        )
        .detuning.uniform.piecewise_linear(
            durations, [-30, -30, "final_detuning", "final_detuning"]
        )
    )

    mis_udg_job = mis_udg_program.batch_assign(final_detuning=np.linspace(0, 80, 81))

    hw_job = mis_udg_job.braket_local_simulator(100).run_async()
"""



from bloqade.builder.base import Builder
from bloqade.builder.field import Rabi, Detuning


class LevelCoupling(Builder):
    @property
    def detuning(
        self,
    ) -> Detuning:  # field is summation of one or more drives,
        # waveform + spatial modulation = drive
        """
        Specify the [`Detuning`][bloqade.builder.field.Detuning]
         [`Field`][bloqade.builder.Field] of your program.

        A "field" is a summation of one or more "drives", with a drive being the sum
        of a waveform and spatial modulation.

        You are currently building the spatial modulation component and will be
        able to specify a waveform.

        - You can do this by:
            - `...detuning.uniform`: To address all atoms in the field
            - `...detuning.location(locations, scales)`: To address atoms at specific
                locations via indices
            - `...detuning.scale(coeffs)`
                - To address all atoms with an individual scale factor

        """

        return Detuning(self)

    @property
    def rabi(self) -> Rabi:
        """
        Specify the complex-valued [`Rabi`][bloqade.builder.field.Rabi]
        field of your program.

        The Rabi field is composed of a real-valued Amplitude and Phase field.

        - Next possible steps to build your program are
          creating the [`RabiAmplitude`][bloqade.builder.field.RabiAmplitude] field
          and [`RabiPhase`][bloqade.builder.field.RabiAmplitude] field of the field:
            - `...rabi.amplitude`: To create the Rabi amplitude field
            - `...rabi.phase`: To create the Rabi phase field

        """

        return Rabi(self)


class Rydberg(LevelCoupling):
    """
    This node represent level coupling of rydberg state.

    Examples:

        - To reach the node from the start node:

        >>> node = bloqade.start.rydberg
        >>> type(node)
        <class 'bloqade.builder.coupling.Rydberg'>

        - Rydberg level coupling have two reachable field nodes:

            - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
            - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

        >>> ryd_detune = bloqade.start.rydberg.detuning
        >>> ryd_rabi = bloqade.start.rydberg.rabi

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.sequence import rydberg

        return rydberg


class Hyperfine(LevelCoupling):
    """
    This node represent level coupling between hyperfine state.

    Examples:

        - To reach the node from the start node:

        >>> node = bloqade.start.hyperfine
        >>> type(node)
        <class 'bloqade.builder.coupling.Hyperfine'>

        - Hyperfine level coupling have two reachable field nodes:

            - detuning term (See also [`Detuning`][bloqade.builder.field.Detuning])
            - rabi term (See also [`Rabi`][bloqade.builder.field.Rabi])

        >>> hyp_detune = bloqade.start.hyperfine.detuning
        >>> hyp_rabi = bloqade.start.hyperfine.rabi

    """

    def __bloqade_ir__(self):
        from bloqade.ir.control.sequence import hyperfine

        return hyperfine



import bloqade.ir.location as location
from bloqade import start, load, save

from bloqade.ir import Linear, Constant
import numpy as np


prog = (
    location.Square(6)
    .rydberg.detuning.uniform.apply(
        Constant("initial_detuning", "up_time")
        .append(Linear("initial_detuning", "final_detuning", "anneal_time"))
        .append(Constant("final_detuning", "up_time"))
    )
    .rabi.amplitude.uniform.apply(
        Linear(0.0, "rabi_amplitude_max", "up_time")
        .append(Constant("rabi_amplitude_max", "anneal_time"))
        .append(Linear("rabi_amplitude_max", 0.0, "up_time"))
    )
    .assign(initial_detuning=-10, up_time=0.1, anneal_time=3.8, rabi_amplitude_max=15)
    .batch_assign(final_detuning=np.linspace(0, 10, 5))
)
task = prog.quera.mock()


# future = task.run_async(shots=100)  ## non0-blk

# future.fetch()

# assert len(future.tasks) == 5

# print(future.report())


def test_serializer():
    batch = (
        start.add_position((0, 0))
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 0.5, 0.1], [1.0, 2.0, 3.0, 4.0]
        )
        .constant(4.0, 1)
        .braket.local_emulator()
        .run(1)
    )

    save(batch, "test.json")
    new_batch = load("test.json")
    assert batch.tasks == new_batch.tasks
